import pyteal as pt


def test_constructor():
    block1 = pt.TealConditionalBlock([])
    assert block1.ops == []
    assert block1.trueBlock is None
    assert block1.falseBlock is None

    block2 = pt.TealConditionalBlock([pt.TealOp(None, pt.Op.int, 1)])
    assert block2.ops == [pt.TealOp(None, pt.Op.int, 1)]
    assert block2.trueBlock is None
    assert block2.falseBlock is None


def test_true_block():
    block = pt.TealConditionalBlock([])
    block.setTrueBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)]))
    assert block.trueBlock == pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)])
    assert block.getOutgoing() == [
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)])
    ]


def test_false_block():
    block = pt.TealConditionalBlock([])
    block.setFalseBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)]))
    assert block.falseBlock == pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)])


def test_outgoing():
    emptyBlock = pt.TealConditionalBlock([])
    assert emptyBlock.getOutgoing() == []

    trueBlock = pt.TealConditionalBlock([])
    trueBlock.setTrueBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')]))
    assert trueBlock.getOutgoing() == [
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')])
    ]

    falseBlock = pt.TealConditionalBlock([])
    falseBlock.setFalseBlock(
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    )
    assert falseBlock.getOutgoing() == [
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    ]

    bothBlock = pt.TealConditionalBlock([])
    bothBlock.setTrueBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')]))
    bothBlock.setFalseBlock(
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')])
    )
    assert bothBlock.getOutgoing() == [
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"true"')]),
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"false"')]),
    ]



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_for_compiles():
    i = pt.ScratchVar()

    expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1))).Do(
        pt.App.globalPut(pt.Itob(pt.Int(0)), pt.Itob(pt.Int(2)))
    )
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()
    expr.__teal__(options)


def test_nested_for_compiles():
    i = pt.ScratchVar()
    expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1))).Do(
        pt.Seq(
            [
                pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1))).Do(
                    pt.Seq([i.store(pt.Int(0))])
                )
            ]
        )
    )
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()


def test_continue_break():
    i = pt.ScratchVar()
    expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1))).Do(
        pt.Seq([pt.If(pt.Int(1), pt.Break(), pt.Continue())])
    )
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()
    expr.__teal__(options)


def test_for():
    i = pt.ScratchVar()
    items = [
        (i.store(pt.Int(0))),
        i.load() < pt.Int(10),
        i.store(i.load() + pt.Int(1)),
        pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2)),
    ]
    expr = pt.For(items[0], items[1], items[2]).Do(pt.Seq([items[3]]))

    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected, varEnd = items[0].__teal__(options)
    condStart, condEnd = items[1].__teal__(options)
    stepStart, stepEnd = items[2].__teal__(options)
    do, doEnd = pt.Seq([items[3]]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    varEnd.setNextBlock(condStart)
    doEnd.setNextBlock(stepStart)

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    stepEnd.setNextBlock(condStart)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_for_continue():
    i = pt.ScratchVar()
    items = [
        (i.store(pt.Int(0))),
        i.load() < pt.Int(10),
        i.store(i.load() + pt.Int(1)),
        pt.If(i.load() < pt.Int(4), pt.Continue()),
        pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2)),
    ]
    expr = pt.For(items[0], items[1], items[2]).Do(pt.Seq([items[3], items[4]]))

    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    options.enterLoop()

    expected, varEnd = items[0].__teal__(options)
    condStart, condEnd = items[1].__teal__(options)
    stepStart, stepEnd = items[2].__teal__(options)
    do, doEnd = pt.Seq([items[3], items[4]]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    doEnd.setNextBlock(stepStart)
    stepEnd.setNextBlock(condStart)

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    varEnd.setNextBlock(condStart)

    _, continueBlocks = options.exitLoop()

    for block in continueBlocks:
        block.setNextBlock(stepStart)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_for_break():
    i = pt.ScratchVar()
    items = [
        (i.store(pt.Int(0))),
        i.load() < pt.Int(10),
        i.store(i.load() + pt.Int(1)),
        pt.If(i.load() == pt.Int(6), pt.Break()),
        pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2)),
    ]
    expr = pt.For(items[0], items[1], items[2]).Do(pt.Seq([items[3], items[4]]))

    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    options.enterLoop()

    expected, varEnd = items[0].__teal__(options)
    condStart, condEnd = items[1].__teal__(options)
    stepStart, stepEnd = items[2].__teal__(options)
    do, doEnd = pt.Seq([items[3], items[4]]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    doEnd.setNextBlock(stepStart)
    stepEnd.setNextBlock(condStart)

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    varEnd.setNextBlock(condStart)

    breakBlocks, _ = options.exitLoop()

    for block in breakBlocks:
        block.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_invalid_for():
    with pytest.raises(TypeError):
        expr = pt.For()

    with pytest.raises(TypeError):
        expr = pt.For(pt.Int(2))

    with pytest.raises(TypeError):
        expr = pt.For(pt.Int(1), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        i = pt.ScratchVar()
        expr = pt.For(i.store(pt.Int(0)), pt.Int(1), pt.Int(2))
        expr.__teal__(options)

    with pytest.raises(pt.TealCompileError):
        i = pt.ScratchVar()
        expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1)))
        expr.type_of()

    with pytest.raises(pt.TealCompileError):
        i = pt.ScratchVar()
        expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1)))
        expr.__str__()

    with pytest.raises(pt.TealTypeError):
        i = pt.ScratchVar()
        expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1))).Do(
            pt.Int(0)
        )

    with pytest.raises(pt.TealTypeError):
        i = pt.ScratchVar()
        expr = pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1))).Do(
            pt.Pop(pt.Int(1)), pt.Int(2)
        )

    with pytest.raises(pt.TealCompileError):
        expr = (
            pt.For(i.store(pt.Int(0)), pt.Int(1), i.store(i.load() + pt.Int(1)))
            .Do(pt.Continue())
            .Do(pt.Continue())
        )
        expr.__str__()


def test_for_multi():
    i = pt.ScratchVar()
    items = [
        (i.store(pt.Int(0))),
        i.load() < pt.Int(10),
        i.store(i.load() + pt.Int(1)),
        [pt.Pop(pt.Int(1)), pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2))],
    ]
    expr = pt.For(items[0], items[1], items[2]).Do(*items[3])

    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected, varEnd = items[0].__teal__(options)
    condStart, condEnd = items[1].__teal__(options)
    stepStart, stepEnd = items[2].__teal__(options)
    do, doEnd = pt.Seq(items[3]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    varEnd.setNextBlock(condStart)
    doEnd.setNextBlock(stepStart)

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    stepEnd.setNextBlock(condStart)

    actual, _ = expr.__teal__(options)

    assert actual == expected



from typing import TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.ir import TealOp, Op, TealBlock, TealSimpleBlock, TealConditionalBlock
from pyteal.ast.expr import Expr
from pyteal.ast.comment import Comment
from pyteal.ast.seq import Seq

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Assert(Expr):
    """A control flow expression to verify that a condition is true."""

    def __init__(
        self, cond: Expr, *additional_conds: Expr, comment: str = None
    ) -> None:
        """Create an assert statement that raises an error if the condition is false.

        Args:
            cond: The condition to check. Must evaluate to a uint64.
            *additional_conds: Additional conditions to check. Must evaluate to uint64.
            comment: String comment to place on the line immediately prior to the assert op
        """
        super().__init__()
        require_type(cond, TealType.uint64)
        for cond_single in additional_conds:
            require_type(cond_single, TealType.uint64)

        self.comment = comment
        self.cond = [cond] + list(additional_conds)

    def __teal__(self, options: "CompileOptions"):
        if len(self.cond) > 1:
            asserts: list[Expr] = []
            for cond in self.cond:
                asrt = Assert(cond, comment=self.comment)
                asrt.trace = cond.trace
                asserts.append(asrt)
            return Seq(*asserts).__teal__(options)

        if options.version >= Op.assert_.min_version:
            # use assert op if available
            conds: list[Expr] = [self.cond[0]]
            if self.comment is not None:
                conds.append(Comment(self.comment))
            return TealBlock.FromOp(options, TealOp(self, Op.assert_), *conds)

        # if assert op is not available, use branches and err
        condStart, condEnd = self.cond[0].__teal__(options)

        end = TealSimpleBlock([])
        errBlock = TealSimpleBlock([TealOp(self, Op.err)])

        branchBlock = TealConditionalBlock([])
        branchBlock.setTrueBlock(end)
        branchBlock.setFalseBlock(errBlock)

        condEnd.setNextBlock(branchBlock)

        return condStart, end

    def __str__(self):
        return "(Assert {})".format(self.cond)

    def type_of(self):
        return TealType.none

    def has_return(self):
        return False


Assert.__module__ = "pyteal"



import pytest

import pyteal as pt

avm2Options = pt.CompileOptions(version=2)
avm3Options = pt.CompileOptions(version=3)


def test_teal_2_assert():
    arg = pt.Int(1)
    expr = pt.Assert(arg)
    assert expr.type_of() == pt.TealType.none

    expected, _ = arg.__teal__(avm2Options)
    expectedBranch = pt.TealConditionalBlock([])
    expectedBranch.setTrueBlock(pt.TealSimpleBlock([]))
    expectedBranch.setFalseBlock(pt.Err().__teal__(avm2Options)[0])
    expected.setNextBlock(expectedBranch)

    actual, _ = expr.__teal__(avm2Options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_teal_2_assert_multi():
    args = [pt.Int(1), pt.Int(2)]
    expr = pt.Assert(*args)
    assert expr.type_of() == pt.TealType.none

    firstAssert = pt.Assert(args[0])
    secondAssert = pt.Assert(args[1])

    expected, _ = pt.Seq(firstAssert, secondAssert).__teal__(avm2Options)

    actual, _ = expr.__teal__(avm2Options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_teal_3_assert():
    arg = pt.Int(1)
    expr = pt.Assert(arg)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg, pt.Op.int, 1), pt.TealOp(expr, pt.Op.assert_)]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_teal_3_assert_multi():
    args = [pt.Int(1), pt.Int(2)]
    expr = pt.Assert(*args)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [pt.TealOp(args[0], pt.Op.int, 1), pt.TealOp(expr, pt.Op.assert_)]
        + [pt.TealOp(args[1], pt.Op.int, 2), pt.TealOp(expr, pt.Op.assert_)]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_assert_comment():
    comment = "Make sure 1 is true"
    expr = pt.Assert(pt.Int(1), comment=comment)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.comment, comment),
            pt.TealOp(None, pt.Op.assert_),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_assert_comment_multi():
    comment = "Make sure numbers > 0 are true"
    expr = pt.Assert(pt.Int(1), pt.Int(2), pt.Int(3), comment=comment)
    assert expr.type_of() == pt.TealType.none

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.comment, comment),
            pt.TealOp(None, pt.Op.assert_),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.comment, comment),
            pt.TealOp(None, pt.Op.assert_),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.comment, comment),
            pt.TealOp(None, pt.Op.assert_),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_assert_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Assert(pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Assert(pt.Int(1), pt.Txn.receiver())



from typing import cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int
from pyteal.ast.ternaryexpr import TernaryExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class ReplaceExpr(Expr):
    """An expression for replacing a section of a byte string at a given start index"""

    def __init__(self, original: Expr, start: Expr, replacement: Expr) -> None:
        super().__init__()

        require_type(original, TealType.bytes)
        require_type(start, TealType.uint64)
        require_type(replacement, TealType.bytes)

        self.original = original
        self.start = start
        self.replacement = replacement

    # helper method for correctly populating op
    def __get_op(self, options: "CompileOptions"):
        s = cast(Int, self.start).value
        if s < 2**8:
            return Op.replace2
        else:
            return Op.replace3

    def __teal__(self, options: "CompileOptions"):
        if not isinstance(self.start, Int):
            return TernaryExpr(
                Op.replace3,
                (TealType.bytes, TealType.uint64, TealType.bytes),
                TealType.bytes,
                self.original,
                self.start,
                self.replacement,
            ).__teal__(options)

        op = self.__get_op(options)

        verifyProgramVersion(
            op.min_version,
            options.version,
            "Program version too low to use op {}".format(op),
        )

        s = cast(Int, self.start).value
        if op == Op.replace2:
            return TealBlock.FromOp(
                options, TealOp(self, op, s), self.original, self.replacement
            )
        elif op == Op.replace3:
            return TealBlock.FromOp(
                options, TealOp(self, op), self.original, self.start, self.replacement
            )

    def __str__(self):
        return "(Replace {} {} {})".format(self.original, self.start, self.replacement)

    def type_of(self):
        return TealType.bytes

    def has_return(self):
        return False


def Replace(original: Expr, start: Expr, replacement: Expr) -> Expr:
    """
    Replace a portion of original bytes with new bytes at a given starting point.

    Requires program version 7 or higher.

    Args:
        original: The value containing the original bytes. Must evaluate to bytes.
        start: The index of the byte where replacement starts. Must evaluate to an integer less than Len(original).
        replacement: The value containing the replacement bytes. Must evaluate to bytes with length at most Len(original) - start.
    """
    return ReplaceExpr(original, start, replacement)



import re
from typing import List, Dict, Set, Optional, TypeVar
from collections import OrderedDict

from pyteal.errors import TealInputError
from pyteal.types import TealType
from pyteal.ast import SubroutineDefinition
from pyteal.ir import TealComponent, TealOp, Op

# generic type variable
Node = TypeVar("Node")


def graph_search(graph: Dict[Node, Set[Node]], start: Node, end: Node) -> bool:
    """Check whether a path between start and end exists in the graph.

    This works even if start == end, in which case True is only returned if the
    node belongs to a cycle.
    """
    visited: Set[Node] = set()
    stack: List[Node] = list(graph[start])

    while len(stack) != 0:
        current = stack.pop()
        if current in visited:
            continue
        visited.add(current)
        if end == current:
            return True
        stack += list(graph[current])

    return False


def findRecursionPoints(
    subroutineGraph: Dict[SubroutineDefinition, Set[SubroutineDefinition]]
) -> Dict[SubroutineDefinition, Set[SubroutineDefinition]]:
    """Find all subroutine calls which may result in the current subroutine being called again
    recursively.

    Args:
        subroutineGraph: A graph of subroutines. Each key is a subroutine (the main routine should
            be present), which represents a node in the graph. Each value is a set of all
            subroutines that specific subroutine calls, which represent directional edges in the
            graph.

    Returns:
        A dictionary whose keys are the same as subroutineGraph, and whose values are a subset of
        the key's values from subroutineGraph. Each element in this subset represents a subroutine
        which may reenter the calling subroutine.
    """
    reentryPoints: Dict[SubroutineDefinition, Set[SubroutineDefinition]] = dict()

    for subroutine in subroutineGraph.keys():
        # perform a depth first search to see which callers (if any) have a path to invoke the calling subroutine again
        reentryPoints[subroutine] = set(
            callee
            for callee in subroutineGraph[subroutine]
            if graph_search(subroutineGraph, callee, subroutine)
        )

    return reentryPoints


def find_recursive_path(
    subroutine_graph: Dict[SubroutineDefinition, Set[SubroutineDefinition]],
    subroutine: SubroutineDefinition,
) -> List[SubroutineDefinition]:
    visited = set()
    loop = []

    def dfs(x):
        if x in visited:
            return False

        visited.add(x)
        loop.append(x)
        for y in subroutine_graph[x]:
            if y == subroutine:
                loop.append(y)
                return True
            if dfs(y):
                return True
        loop.pop()
        return False

    found = dfs(subroutine)
    return loop if found else []


def spillLocalSlotsDuringRecursion(
    version: int,
    subroutineMapping: Dict[Optional[SubroutineDefinition], List[TealComponent]],
    subroutineGraph: Dict[SubroutineDefinition, Set[SubroutineDefinition]],
    localSlots: Dict[Optional[SubroutineDefinition], Set[int]],
) -> None:
    """In order to prevent recursion from modifying the local scratch slots a subroutine uses,
    subroutines must "spill" their local slots to the stack before calling any other subroutine
    which may invoke the calling subroutine.

    "Spill to stack" means loading all local slots onto the stack, invoking the subroutine which may
    result in recursion, then restoring all local slots from the stack. This prevents the local
    slots from being modifying by a new recursive invocation of the current subroutine.

    Args:
        version: The current program version being assembled.
        subroutineMapping: A dictionary containing a list of TealComponents for every subroutine in
            a program. The key None is taken to indicate the main program routine. This input may be
            modified by this function in order to spill subroutine slots.
        subroutineGraph: A graph of subroutines. Each key is a subroutine (the main routine should
            not be present), which represents a node in the graph. Each value is a set of all
            subroutines that specific subroutine calls, which represent directional edges in the
            graph.
        localSlots: The output from the function `assignScratchSlotsToSubroutines`, which indicates
            the local slots which must be spilled for each subroutine.
    """
    recursivePoints = findRecursionPoints(subroutineGraph)

    recursive_byref = None
    for k, v in recursivePoints.items():
        if v and k.by_ref_args:
            recursive_byref = k
            break

    if recursive_byref:
        msg = "ScratchVar arguments not allowed in recursive subroutines, but a recursive call-path was detected: {}()"
        raise TealInputError(
            msg.format(
                "()-->".join(
                    f.name()
                    for f in find_recursive_path(subroutineGraph, recursive_byref)
                )
            )
        )

    coverAvailable = version >= Op.cover.min_version

    for subroutine, reentryPoints in recursivePoints.items():
        slots = list(sorted(slot for slot in localSlots[subroutine]))

        if len(reentryPoints) == 0 or len(slots) == 0:
            # no need to spill slots
            continue

        ops = subroutineMapping[subroutine]
        newOps: List[TealComponent] = []

        for stmt in ops:
            before: List[TealComponent] = []
            after: List[TealComponent] = []

            calledSubroutines = stmt.getSubroutines()
            # the only opcode that references subroutines is callsub, and it should only ever
            # reference one subroutine at a time
            assert (
                len(calledSubroutines) <= 1
            ), "Multiple subroutines are called from the same TealComponent"

            reentrySubroutineCalls = list(reentryPoints.intersection(calledSubroutines))
            if len(reentrySubroutineCalls) != 0:
                # A subroutine is being called which may reenter the current subroutine, so insert
                # ops to spill local slots to the stack before calling the subroutine and also to
                # restore the local slots after returning from the subroutine. This prevents a
                # reentry into the current subroutine from modifying variables we are currently
                # using.

                # reentrySubroutineCalls should have a length of 1, since calledSubroutines has a
                # maximum length of 1
                reentrySubroutineCall = reentrySubroutineCalls[0]
                numArgs = reentrySubroutineCall.argument_count()

                digArgs = True
                coverSpilledSlots = False
                uncoverArgs = False
                if coverAvailable:
                    digArgs = False
                    if len(slots) < numArgs:
                        coverSpilledSlots = True
                    else:
                        uncoverArgs = True

                for slot in slots:
                    # spill local slots to the stack
                    before.append(TealOp(None, Op.load, slot))

                    if coverSpilledSlots:
                        # numArgs is guaranteed to be at least 2 here (since numArgs > len(slots)
                        # and len(slots) must be at least 1 for the code to get this far), so no
                        # need to replace this with swap if numArgs is 1
                        before.append(TealOp(None, Op.cover, numArgs))

                for _ in range(numArgs):
                    # pull the subroutine arguments to the top of the stack, above the just spilled
                    # local slots, if needed

                    stackDistance = len(slots) + numArgs - 1

                    if uncoverArgs:
                        if stackDistance == 1:
                            before.append(TealOp(None, Op.swap))
                        else:
                            before.append(TealOp(None, Op.uncover, stackDistance))

                    if digArgs:
                        before.append(
                            TealOp(
                                None,
                                Op.dig,
                                stackDistance,
                            )
                        )
                        # because we are stuck using dig instead of uncover in AVM 4, we'll need to
                        # pop all of the dug up arguments after the function returns

                hideReturnValueInFirstSlot = False

                if subroutine.return_type != TealType.none:
                    # if the subroutine returns a value on the stack, we need to preserve this after
                    # restoring all local slots.

                    if len(slots) == 1:
                        after.append(TealOp(None, Op.swap))
                    elif coverAvailable:
                        after.append(TealOp(None, Op.cover, len(slots)))
                    else:
                        # Store the return value into slots[0] temporarily
                        hideReturnValueInFirstSlot = True
                        after.append(TealOp(None, Op.store, slots[0]))

                for slot in slots[::-1]:
                    # restore slots, iterating in reverse because slots[-1] is at the top of the stack
                    if hideReturnValueInFirstSlot and slot is slots[0]:
                        # time to restore the return value to the top of the stack

                        # slots[0] is being used to store the return value, so load it again
                        after.append(TealOp(None, Op.load, slot))

                        # swap the return value with the actual value of slot[0] on the stack
                        after.append(TealOp(None, Op.swap))

                    after.append(TealOp(None, Op.store, slot))

                if digArgs:
                    for _ in range(numArgs):
                        # clear out the duplicate arguments that were dug up previously, since dig
                        # does not pop the dug values -- once we use cover/uncover to properly set up
                        # the spilled slots, this will no longer be necessary
                        if subroutine.return_type != TealType.none:
                            # if there is a return value on top of the stack, we need to preserve
                            # it, so swap it with the subroutine argument that's below it on the
                            # stack
                            after.append(TealOp(None, Op.swap))
                        after.append(TealOp(None, Op.pop))

            newOps += before
            newOps.append(stmt)
            newOps += after

        subroutineMapping[subroutine] = newOps


def resolveSubroutines(
    subroutineMapping: Dict[Optional[SubroutineDefinition], List[TealComponent]]
) -> Dict[SubroutineDefinition, str]:
    """Resolve referenced subroutines for an entire program.

    Args:
        subroutineMapping: A dictionary containing a list of TealComponents for every subroutine in
            a program. The key None is taken to indicate the main program routine. This input will
            be modified by this function in order to assign labels to subroutines.

    Returns:
        An ordered dictionary whose keys are the same as subroutineMapping, minus the None key. The
        values of this dictionary will be the resolved label for each subroutine. The order of this
        dictionary is taken to be the official ordering of the subroutines, which should be used in
        later code generation steps.
    """
    allButMainRoutine = (
        subroutine for subroutine in subroutineMapping.keys() if subroutine is not None
    )

    subroutineOrder = sorted(allButMainRoutine, key=lambda subroutine: subroutine.id)
    subroutineToLabel: Dict[SubroutineDefinition, str] = OrderedDict()
    for index, subroutine in enumerate(subroutineOrder):
        safer_name = re.sub(r"[^A-Za-z0-9]", "", subroutine.name())
        subroutineToLabel[subroutine] = "{}_{}".format(safer_name, index)

    for subroutine, label in subroutineToLabel.items():
        for ops in subroutineMapping.values():
            for stmt in ops:
                stmt.resolveSubroutine(subroutine, label)

    return subroutineToLabel



from typing import NamedTuple, List, Callable, Literal
import pytest

import pyteal as pt
from pyteal import abi
from pyteal.ast.abi.tuple import _encode_tuple, _index_tuple, TupleElement
from pyteal.ast.abi.bool import _encode_bool_sequence
from pyteal.ast.abi.util import substring_for_decoding
from pyteal.ast.abi.type_test import ContainerType

options = pt.CompileOptions(version=5)


def test_encodeTuple():
    class EncodeTest(NamedTuple):
        types: List[abi.BaseType]
        expected: pt.Expr

    # variables used to construct the tests
    uint64_a = abi.Uint64()
    uint64_b = abi.Uint64()
    uint16_a = abi.Uint16()
    uint16_b = abi.Uint16()
    bool_a = abi.Bool()
    bool_b = abi.Bool()
    tuple_a = abi.Tuple(abi.TupleTypeSpec(abi.BoolTypeSpec(), abi.BoolTypeSpec()))
    dynamic_array_a = abi.DynamicArray(abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec()))
    dynamic_array_b = abi.DynamicArray(abi.DynamicArrayTypeSpec(abi.Uint16TypeSpec()))
    dynamic_array_c = abi.DynamicArray(abi.DynamicArrayTypeSpec(abi.BoolTypeSpec()))
    tail_holder = pt.ScratchVar()
    encoded_tail = pt.ScratchVar()

    tests: List[EncodeTest] = [
        EncodeTest(types=[], expected=pt.Bytes("")),
        EncodeTest(types=[uint64_a], expected=uint64_a.encode()),
        EncodeTest(
            types=[uint64_a, uint64_b],
            expected=pt.Concat(uint64_a.encode(), uint64_b.encode()),
        ),
        EncodeTest(types=[bool_a], expected=bool_a.encode()),
        EncodeTest(
            types=[bool_a, bool_b], expected=_encode_bool_sequence([bool_a, bool_b])
        ),
        EncodeTest(
            types=[bool_a, bool_b, uint64_a],
            expected=pt.Concat(
                _encode_bool_sequence([bool_a, bool_b]), uint64_a.encode()
            ),
        ),
        EncodeTest(
            types=[uint64_a, bool_a, bool_b],
            expected=pt.Concat(
                uint64_a.encode(), _encode_bool_sequence([bool_a, bool_b])
            ),
        ),
        EncodeTest(
            types=[uint64_a, bool_a, bool_b, uint64_b],
            expected=pt.Concat(
                uint64_a.encode(),
                _encode_bool_sequence([bool_a, bool_b]),
                uint64_b.encode(),
            ),
        ),
        EncodeTest(
            types=[uint64_a, bool_a, uint64_b, bool_b],
            expected=pt.Concat(
                uint64_a.encode(), bool_a.encode(), uint64_b.encode(), bool_b.encode()
            ),
        ),
        EncodeTest(types=[tuple_a], expected=tuple_a.encode()),
        EncodeTest(
            types=[uint64_a, tuple_a, bool_a, bool_b],
            expected=pt.Concat(
                uint64_a.encode(),
                tuple_a.encode(),
                _encode_bool_sequence([bool_a, bool_b]),
            ),
        ),
        EncodeTest(
            types=[dynamic_array_a],
            expected=pt.Concat(
                pt.Seq(
                    encoded_tail.store(dynamic_array_a.encode()),
                    tail_holder.store(encoded_tail.load()),
                    uint16_a.set(2),
                    uint16_a.encode(),
                ),
                tail_holder.load(),
            ),
        ),
        EncodeTest(
            types=[uint64_a, dynamic_array_a],
            expected=pt.Concat(
                uint64_a.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_a.encode()),
                    tail_holder.store(encoded_tail.load()),
                    uint16_a.set(8 + 2),
                    uint16_a.encode(),
                ),
                tail_holder.load(),
            ),
        ),
        EncodeTest(
            types=[uint64_a, dynamic_array_a, uint64_b],
            expected=pt.Concat(
                uint64_a.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_a.encode()),
                    tail_holder.store(encoded_tail.load()),
                    uint16_a.set(8 + 2 + 8),
                    uint16_a.encode(),
                ),
                uint64_b.encode(),
                tail_holder.load(),
            ),
        ),
        EncodeTest(
            types=[uint64_a, dynamic_array_a, bool_a, bool_b],
            expected=pt.Concat(
                uint64_a.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_a.encode()),
                    tail_holder.store(encoded_tail.load()),
                    uint16_a.set(8 + 2 + 1),
                    uint16_a.encode(),
                ),
                _encode_bool_sequence([bool_a, bool_b]),
                tail_holder.load(),
            ),
        ),
        EncodeTest(
            types=[uint64_a, dynamic_array_a, uint64_b, dynamic_array_b],
            expected=pt.Concat(
                uint64_a.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_a.encode()),
                    tail_holder.store(encoded_tail.load()),
                    uint16_a.set(8 + 2 + 8 + 2),
                    uint16_b.set(uint16_a.get() + pt.Len(encoded_tail.load())),
                    uint16_a.encode(),
                ),
                uint64_b.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_b.encode()),
                    tail_holder.store(
                        pt.Concat(tail_holder.load(), encoded_tail.load())
                    ),
                    uint16_a.set(uint16_b),
                    uint16_a.encode(),
                ),
                tail_holder.load(),
            ),
        ),
        EncodeTest(
            types=[
                uint64_a,
                dynamic_array_a,
                uint64_b,
                dynamic_array_b,
                bool_a,
                bool_b,
                dynamic_array_c,
            ],
            expected=pt.Concat(
                uint64_a.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_a.encode()),
                    tail_holder.store(encoded_tail.load()),
                    uint16_a.set(8 + 2 + 8 + 2 + 1 + 2),
                    uint16_b.set(uint16_a.get() + pt.Len(encoded_tail.load())),
                    uint16_a.encode(),
                ),
                uint64_b.encode(),
                pt.Seq(
                    encoded_tail.store(dynamic_array_b.encode()),
                    tail_holder.store(
                        pt.Concat(tail_holder.load(), encoded_tail.load())
                    ),
                    uint16_a.set(uint16_b),
                    uint16_b.set(uint16_a.get() + pt.Len(encoded_tail.load())),
                    uint16_a.encode(),
                ),
                _encode_bool_sequence([bool_a, bool_b]),
                pt.Seq(
                    encoded_tail.store(dynamic_array_c.encode()),
                    tail_holder.store(
                        pt.Concat(tail_holder.load(), encoded_tail.load())
                    ),
                    uint16_a.set(uint16_b),
                    uint16_a.encode(),
                ),
                tail_holder.load(),
            ),
        ),
    ]

    for i, test in enumerate(tests):
        expr = _encode_tuple(test.types)
        assert expr.type_of() == pt.TealType.bytes
        assert not expr.has_return()

        expected, _ = test.expected.__teal__(options)
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        if any(t.type_spec().is_dynamic() for t in test.types):
            with pt.TealComponent.Context.ignoreExprEquality():
                with pt.TealComponent.Context.ignoreScratchSlotEquality():
                    assert actual == expected, "Test at index {} failed".format(i)

            assert pt.TealBlock.MatchScratchSlotReferences(
                pt.TealBlock.GetReferencedScratchSlots(actual),
                pt.TealBlock.GetReferencedScratchSlots(expected),
            )
            continue

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected, "Test at index {} failed".format(i)


def test_indexTuple():
    class IndexTest(NamedTuple):
        types: List[abi.TypeSpec]
        typeIndex: int
        expected: Callable[[abi.BaseType], pt.Expr]

    # variables used to construct the tests
    uint64_t = abi.Uint64TypeSpec()
    byte_t = abi.ByteTypeSpec()
    bool_t = abi.BoolTypeSpec()
    tuple_t = abi.TupleTypeSpec(abi.BoolTypeSpec(), abi.BoolTypeSpec())
    dynamic_array_t1 = abi.DynamicArrayTypeSpec(abi.Uint64TypeSpec())
    dynamic_array_t2 = abi.DynamicArrayTypeSpec(abi.Uint16TypeSpec())

    encoded = pt.Bytes("encoded")

    tests: List[IndexTest] = [
        IndexTest(
            types=[uint64_t],
            typeIndex=0,
            expected=lambda output: output.decode(encoded),
        ),
        IndexTest(
            types=[uint64_t, uint64_t],
            typeIndex=0,
            expected=lambda output: output.decode(encoded, length=pt.Int(8)),
        ),
        IndexTest(
            types=[uint64_t, uint64_t],
            typeIndex=1,
            expected=lambda output: output.decode(encoded, start_index=pt.Int(8)),
        ),
        IndexTest(
            types=[uint64_t, byte_t, uint64_t],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded, start_index=pt.Int(8), length=pt.Int(1)
            ),
        ),
        IndexTest(
            types=[uint64_t, byte_t, uint64_t],
            typeIndex=2,
            expected=lambda output: output.decode(
                encoded, start_index=pt.Int(9), length=pt.Int(8)
            ),
        ),
        IndexTest(
            types=[bool_t],
            typeIndex=0,
            expected=lambda output: output.decode_bit(encoded, pt.Int(0)),
        ),
        IndexTest(
            types=[bool_t, bool_t],
            typeIndex=0,
            expected=lambda output: output.decode_bit(encoded, pt.Int(0)),
        ),
        IndexTest(
            types=[bool_t, bool_t],
            typeIndex=1,
            expected=lambda output: output.decode_bit(encoded, pt.Int(1)),
        ),
        IndexTest(
            types=[uint64_t, bool_t],
            typeIndex=1,
            expected=lambda output: output.decode_bit(encoded, pt.Int(8 * 8)),
        ),
        IndexTest(
            types=[uint64_t, bool_t, bool_t],
            typeIndex=1,
            expected=lambda output: output.decode_bit(encoded, pt.Int(8 * 8)),
        ),
        IndexTest(
            types=[uint64_t, bool_t, bool_t],
            typeIndex=2,
            expected=lambda output: output.decode_bit(encoded, pt.Int(8 * 8 + 1)),
        ),
        IndexTest(
            types=[bool_t, uint64_t],
            typeIndex=0,
            expected=lambda output: output.decode_bit(encoded, pt.Int(0)),
        ),
        IndexTest(
            types=[bool_t, uint64_t],
            typeIndex=1,
            expected=lambda output: output.decode(encoded, start_index=pt.Int(1)),
        ),
        IndexTest(
            types=[bool_t, bool_t, uint64_t],
            typeIndex=0,
            expected=lambda output: output.decode_bit(encoded, pt.Int(0)),
        ),
        IndexTest(
            types=[bool_t, bool_t, uint64_t],
            typeIndex=1,
            expected=lambda output: output.decode_bit(encoded, pt.Int(1)),
        ),
        IndexTest(
            types=[bool_t, bool_t, uint64_t],
            typeIndex=2,
            expected=lambda output: output.decode(encoded, start_index=pt.Int(1)),
        ),
        IndexTest(
            types=[tuple_t], typeIndex=0, expected=lambda output: output.decode(encoded)
        ),
        IndexTest(
            types=[byte_t, tuple_t],
            typeIndex=1,
            expected=lambda output: output.decode(encoded, start_index=pt.Int(1)),
        ),
        IndexTest(
            types=[tuple_t, byte_t],
            typeIndex=0,
            expected=lambda output: output.decode(
                encoded,
                start_index=pt.Int(0),
                length=pt.Int(tuple_t.byte_length_static()),
            ),
        ),
        IndexTest(
            types=[byte_t, tuple_t, byte_t],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded,
                start_index=pt.Int(1),
                length=pt.Int(tuple_t.byte_length_static()),
            ),
        ),
        IndexTest(
            types=[dynamic_array_t1],
            typeIndex=0,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(0))
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(1))
            ),
        ),
        IndexTest(
            types=[dynamic_array_t1, byte_t],
            typeIndex=0,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(0))
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1, byte_t],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(1))
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1, byte_t, dynamic_array_t2],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded,
                start_index=pt.ExtractUint16(encoded, pt.Int(1)),
                end_index=pt.ExtractUint16(encoded, pt.Int(4)),
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1, byte_t, dynamic_array_t2],
            typeIndex=3,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(4))
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1, tuple_t, dynamic_array_t2],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded,
                start_index=pt.ExtractUint16(encoded, pt.Int(1)),
                end_index=pt.ExtractUint16(encoded, pt.Int(4)),
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1, tuple_t, dynamic_array_t2],
            typeIndex=3,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(4))
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t2, bool_t, bool_t, dynamic_array_t2],
            typeIndex=1,
            expected=lambda output: output.decode(
                encoded,
                start_index=pt.ExtractUint16(encoded, pt.Int(1)),
                end_index=pt.ExtractUint16(encoded, pt.Int(4)),
            ),
        ),
        IndexTest(
            types=[byte_t, dynamic_array_t1, bool_t, bool_t, dynamic_array_t2],
            typeIndex=4,
            expected=lambda output: output.decode(
                encoded, start_index=pt.ExtractUint16(encoded, pt.Int(4))
            ),
        ),
    ]

    for i, test in enumerate(tests):
        output = test.types[test.typeIndex].new_instance()
        expr = _index_tuple(test.types, encoded, test.typeIndex, output)
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected, _ = test.expected(output).__teal__(options)
        expected.addIncoming()
        expected = pt.TealBlock.NormalizeBlocks(expected)

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected, "Test at index {} failed".format(i)

        with pytest.raises(ValueError):
            _index_tuple(test.types, encoded, len(test.types), output)

        with pytest.raises(ValueError):
            _index_tuple(test.types, encoded, -1, output)

        otherType = abi.Uint64()
        if output.type_spec() == otherType.type_spec():
            otherType = abi.Uint16()

        with pytest.raises(TypeError):
            _index_tuple(test.types, encoded, test.typeIndex, otherType)


def test_TupleTypeSpec_eq():
    tupleA = abi.TupleTypeSpec(
        abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
    )
    tupleB = abi.TupleTypeSpec(
        abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
    )
    tupleC = abi.TupleTypeSpec(
        abi.BoolTypeSpec(), abi.Uint64TypeSpec(), abi.Uint32TypeSpec()
    )
    assert tupleA == tupleA
    assert tupleA == tupleB
    assert tupleA != tupleC


def test_TupleTypeSpec_value_type_specs():
    assert abi.TupleTypeSpec(
        abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
    ).value_type_specs() == [
        abi.Uint64TypeSpec(),
        abi.Uint32TypeSpec(),
        abi.BoolTypeSpec(),
    ]


def test_TupleTypeSpec_length_static():
    tests: List[List[abi.TypeSpec]] = [
        [],
        [abi.Uint64TypeSpec()],
        [
            abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.Uint64TypeSpec()),
            abi.Uint64TypeSpec(),
        ],
        [abi.BoolTypeSpec()] * 8,
    ]

    for i, test in enumerate(tests):
        actual = abi.TupleTypeSpec(*test).length_static()
        expected = len(test)
        assert actual == expected, "Test at index {} failed".format(i)


def test_TupleTypeSpec_new_instance():
    assert isinstance(
        abi.TupleTypeSpec(
            abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
        ).new_instance(),
        abi.Tuple,
    )


def test_TupleTypeSpec_is_dynamic():
    assert not abi.TupleTypeSpec().is_dynamic()
    assert not abi.TupleTypeSpec(
        abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
    ).is_dynamic()
    assert abi.TupleTypeSpec(
        abi.Uint16TypeSpec(), abi.DynamicArrayTypeSpec(abi.Uint8TypeSpec())
    ).is_dynamic()


def test_TupleTypeSpec_str():
    assert str(abi.TupleTypeSpec()) == "()"
    assert str(abi.TupleTypeSpec(abi.TupleTypeSpec())) == "(())"
    assert str(abi.TupleTypeSpec(abi.TupleTypeSpec(), abi.TupleTypeSpec())) == "((),())"
    assert (
        str(
            abi.TupleTypeSpec(
                abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
            )
        )
        == "(uint64,uint32,bool)"
    )
    assert (
        str(
            abi.TupleTypeSpec(
                abi.BoolTypeSpec(), abi.Uint64TypeSpec(), abi.Uint32TypeSpec()
            )
        )
        == "(bool,uint64,uint32)"
    )
    assert (
        str(
            abi.TupleTypeSpec(
                abi.Uint16TypeSpec(), abi.DynamicArrayTypeSpec(abi.Uint8TypeSpec())
            )
        )
        == "(uint16,uint8[])"
    )


def test_TupleTypeSpec_byte_length_static():
    assert abi.TupleTypeSpec().byte_length_static() == 0
    assert abi.TupleTypeSpec(abi.TupleTypeSpec()).byte_length_static() == 0
    assert (
        abi.TupleTypeSpec(abi.TupleTypeSpec(), abi.TupleTypeSpec()).byte_length_static()
        == 0
    )
    assert (
        abi.TupleTypeSpec(
            abi.Uint64TypeSpec(), abi.Uint32TypeSpec(), abi.BoolTypeSpec()
        ).byte_length_static()
        == 8 + 4 + 1
    )
    assert (
        abi.TupleTypeSpec(
            abi.Uint64TypeSpec(),
            abi.Uint32TypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
        ).byte_length_static()
        == 8 + 4 + 1
    )
    assert (
        abi.TupleTypeSpec(
            abi.Uint64TypeSpec(),
            abi.Uint32TypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
            abi.BoolTypeSpec(),
        ).byte_length_static()
        == 8 + 4 + 2
    )

    with pytest.raises(ValueError):
        abi.TupleTypeSpec(
            abi.Uint16TypeSpec(), abi.DynamicArrayTypeSpec(abi.Uint8TypeSpec())
        ).byte_length_static()


def test_Tuple_decode():
    encoded = pt.Bytes("encoded")
    tupleValue = abi.Tuple(abi.TupleTypeSpec(abi.Uint64TypeSpec()))
    for start_index in (None, pt.Int(1)):
        for end_index in (None, pt.Int(2)):
            for length in (None, pt.Int(3)):
                if end_index is not None and length is not None:
                    with pytest.raises(pt.TealInputError):
                        tupleValue.decode(
                            encoded,
                            start_index=start_index,
                            end_index=end_index,
                            length=length,
                        )
                    continue

                expr = tupleValue.decode(
                    encoded, start_index=start_index, end_index=end_index, length=length
                )
                assert expr.type_of() == pt.TealType.none
                assert not expr.has_return()

                expectedExpr = tupleValue.stored_value.store(
                    substring_for_decoding(
                        encoded,
                        start_index=start_index,
                        end_index=end_index,
                        length=length,
                    )
                )
                expected, _ = expectedExpr.__teal__(options)
                expected.addIncoming()
                expected = pt.TealBlock.NormalizeBlocks(expected)

                actual, _ = expr.__teal__(options)
                actual.addIncoming()
                actual = pt.TealBlock.NormalizeBlocks(actual)

                with pt.TealComponent.Context.ignoreExprEquality():
                    assert actual == expected


def test_Tuple_set():
    tupleValue = abi.Tuple(
        abi.TupleTypeSpec(
            abi.Uint8TypeSpec(), abi.Uint16TypeSpec(), abi.Uint32TypeSpec()
        )
    )
    uint8 = abi.Uint8()
    uint16 = abi.Uint16()
    uint32 = abi.Uint32()

    with pytest.raises(pt.TealInputError):
        tupleValue.set()

    with pytest.raises(pt.TealInputError):
        tupleValue.set(uint8, uint16)

    with pytest.raises(pt.TealInputError):
        tupleValue.set(uint8, uint16, uint32, uint32)

    with pytest.raises(pt.TealInputError):
        tupleValue.set(uint8, uint32, uint16)

    with pytest.raises(pt.TealInputError):
        tupleValue.set(uint8, uint16, uint16)

    expr = tupleValue.set(uint8, uint16, uint32)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expectedExpr = tupleValue.stored_value.store(_encode_tuple([uint8, uint16, uint32]))
    expected, _ = expectedExpr.__teal__(options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_Tuple_set_Computed():
    tupleValue = abi.Tuple(
        abi.TupleTypeSpec(
            abi.Uint8TypeSpec(), abi.Uint16TypeSpec(), abi.Uint32TypeSpec()
        )
    )
    computed = ContainerType(
        tupleValue.type_spec(), pt.Bytes("internal representation")
    )
    expr = tupleValue.set(computed)
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(None, pt.Op.byte, '"internal representation"'),
            pt.TealOp(None, pt.Op.store, tupleValue.stored_value.slot),
        ]
    )
    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        tupleValue.set(computed, computed)

    with pytest.raises(pt.TealInputError):
        tupleValue.set(
            ContainerType(abi.TupleTypeSpec(abi.ByteTypeSpec()), pt.Bytes(b"a"))
        )


def test_Tuple_encode():
    tupleValue = abi.Tuple(abi.TupleTypeSpec(abi.Uint64TypeSpec()))
    expr = tupleValue.encode()
    assert expr.type_of() == pt.TealType.bytes
    assert not expr.has_return()

    expected = pt.TealSimpleBlock(
        [pt.TealOp(None, pt.Op.load, tupleValue.stored_value.slot)]
    )

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_Tuple_length():
    tests: List[List[abi.TypeSpec]] = [
        [],
        [abi.Uint64TypeSpec()],
        [
            abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.Uint64TypeSpec()),
            abi.Uint64TypeSpec(),
        ],
        [abi.BoolTypeSpec()] * 8,
    ]

    for i, test in enumerate(tests):
        tupleValue = abi.Tuple(abi.TupleTypeSpec(*test))
        expr = tupleValue.length()
        assert expr.type_of() == pt.TealType.uint64
        assert not expr.has_return()

        expectedLength = len(test)
        expected = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, expectedLength)])

        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected, "Test at index {} failed".format(i)


def test_Tuple_getitem():
    tests: List[List[abi.TypeSpec]] = [
        [],
        [abi.Uint64TypeSpec()],
        [
            abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.Uint64TypeSpec()),
            abi.Uint64TypeSpec(),
        ],
        [abi.BoolTypeSpec()] * 8,
    ]

    for i, test in enumerate(tests):
        tupleValue = abi.Tuple(abi.TupleTypeSpec(*test))
        for j in range(len(test)):
            element = tupleValue[j]
            assert type(element) is TupleElement, "Test at index {} failed".format(i)
            assert element.tuple is tupleValue, "Test at index {} failed".format(i)
            assert element.index == j, "Test at index {} failed".format(i)

        with pytest.raises(pt.TealInputError):
            tupleValue[-1]

        with pytest.raises(pt.TealInputError):
            tupleValue[len(test)]


def test_TupleElement_store_into():
    tests: List[List[abi.TypeSpec]] = [
        [],
        [abi.Uint64TypeSpec()],
        [
            abi.TupleTypeSpec(abi.Uint64TypeSpec(), abi.Uint64TypeSpec()),
            abi.Uint64TypeSpec(),
        ],
        [abi.BoolTypeSpec()] * 8,
    ]

    for i, test in enumerate(tests):
        tupleValue = abi.Tuple(abi.TupleTypeSpec(*test))
        for j in range(len(test)):
            element = TupleElement(tupleValue, j)
            output = test[j].new_instance()

            expr = element.store_into(output)
            assert expr.type_of() == pt.TealType.none
            assert not expr.has_return()

            expectedExpr = _index_tuple(test, tupleValue.encode(), j, output)
            expected, _ = expectedExpr.__teal__(options)
            expected.addIncoming()
            expected = pt.TealBlock.NormalizeBlocks(expected)

            actual, _ = expr.__teal__(options)
            actual.addIncoming()
            actual = pt.TealBlock.NormalizeBlocks(actual)

            with pt.TealComponent.Context.ignoreExprEquality():
                assert actual == expected, "Test at index {} failed".format(i)


def test_NamedTuple_init():
    with pytest.raises(pt.TealInputError, match=r"NamedTuple must be subclassed$"):
        abi.NamedTuple()

    class Empty(abi.NamedTuple):
        pass

    with pytest.raises(
        pt.TealInputError, match=r"Expected fields to be declared but found none$"
    ):
        Empty()

    class ValidField(abi.NamedTuple):
        name: abi.Field[abi.Uint16]

    ValidField()

    class NoField(abi.NamedTuple):
        name: abi.Uint16

    with pytest.raises(
        pt.TealInputError,
        match=r'Type annotation for attribute "name" must be a Field. Got ',
    ):
        NoField()


class NT_0(abi.NamedTuple):
    f0: abi.Field[abi.Uint64]
    f1: abi.Field[abi.Uint32]
    f2: abi.Field[abi.Uint16]
    f3: abi.Field[abi.Uint8]


class NT_1(abi.NamedTuple):
    f0: abi.Field[abi.StaticArray[abi.Bool, Literal[4]]]
    f1: abi.Field[abi.DynamicArray[abi.String]]
    f2: abi.Field[abi.String]
    f3: abi.Field[abi.Bool]
    f4: abi.Field[abi.Address]
    f5: abi.Field[NT_0]


class NT_2(abi.NamedTuple):
    f0: abi.Field[abi.Bool]
    f1: abi.Field[abi.Bool]
    f2: abi.Field[abi.Bool]
    f3: abi.Field[abi.Bool]
    f4: abi.Field[abi.Bool]
    f5: abi.Field[abi.Bool]
    f6: abi.Field[abi.Bool]
    f7: abi.Field[abi.Bool]
    f8: abi.Field[NT_1]


class NT_3(abi.NamedTuple):
    f0: abi.Field[NT_0]
    f1: abi.Field[NT_1]
    f2: abi.Field[NT_2]


@pytest.mark.parametrize("test_case", [NT_0, NT_1, NT_2, NT_3])
def test_NamedTuple_getitem(test_case: type[abi.NamedTuple]):
    tuple_value = test_case()
    tuple_len_static = tuple_value.type_spec().length_static()
    for i in range(tuple_len_static):
        elem_by_field: abi.TupleElement = getattr(tuple_value, f"f{i}")
        elem_by_index: abi.TupleElement = tuple_value[i]

        assert (
            type(elem_by_field) is abi.TupleElement
        ), f"Test case {test_case} at field f{i} must be TupleElement"
        assert (
            type(elem_by_index) is abi.TupleElement
        ), f"Test case {test_case} at index {i} must be TupleElement"

        assert (
            elem_by_field.index == i
        ), f"Test case {test_case} at field f{i} should have index {i}."
        assert (
            elem_by_index.index == i
        ), f"Test case {test_case} at index {i} should have index {i}."

        assert (
            elem_by_field.tuple is tuple_value
        ), f"Test case {test_case} at field f{i} should have attr tuple == {test_case}."
        assert (
            elem_by_index.tuple is tuple_value
        ), f"Test case {test_case} at index {i} should have attr tuple == {test_case}."

        assert (
            elem_by_field.produced_type_spec() == elem_by_index.produced_type_spec()
        ), f"Test case {test_case} at field f{i} type spec unmatching: {elem_by_field.produced_type_spec()} != {elem_by_index.produced_type_spec()}."

    with pytest.raises(KeyError):
        tuple_value.aaaaa

    with pytest.raises(pt.TealInputError):
        tuple_value.f0 = abi.Uint64()


def test_NamedTupleTypeSpec():
    from pyteal.ast.abi.util import type_spec_is_assignable_to

    class Point(abi.NamedTuple):
        x: abi.Field[abi.Uint64]
        y: abi.Field[abi.Uint64]

    class AccountRecord(abi.NamedTuple):
        algoBalance: abi.Field[abi.Uint64]
        assetBalance: abi.Field[abi.Uint64]

    p = Point()
    ar = AccountRecord()

    assert p.type_spec() == p.type_spec()
    assert ar.type_spec() == ar.type_spec()
    assert p.type_spec() != ar.type_spec()
    assert not type_spec_is_assignable_to(p.type_spec(), ar.type_spec())
    assert not type_spec_is_assignable_to(ar.type_spec(), p.type_spec())



import pytest

from pyteal.ast.methodsig import MethodSignature

import pyteal as pt


def test_method():
    expr = MethodSignature("add(uint64,uint64)uint64")
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [pt.TealOp(expr, pt.Op.method_signature, '"add(uint64,uint64)uint64"')]
    )
    actual, _ = expr.__teal__(pt.CompileOptions())
    assert expected == actual


def test_method_invalid():
    with pytest.raises(pt.TealInputError):
        MethodSignature(114514)

    with pytest.raises(pt.TealInputError):
        MethodSignature(['"m0()void"', '"m1()uint64"'])

    with pytest.raises(pt.TealInputError):
        MethodSignature("")



import pytest

import pyteal as pt

avm3Options = pt.CompileOptions(version=3)
avm4Options = pt.CompileOptions(version=4)


def test_gaid_teal_3():
    with pytest.raises(pt.TealInputError):
        pt.GeneratedID(0).__teal__(avm3Options)


def test_gaid():
    expr = pt.GeneratedID(0)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.gaid, 0)])

    actual, _ = expr.__teal__(avm4Options)

    assert actual == expected


def test_gaid_invalid():
    with pytest.raises(pt.TealInputError):
        pt.GeneratedID(-1)

    with pytest.raises(pt.TealInputError):
        pt.GeneratedID(pt.MAX_GROUP_SIZE)


def test_gaid_dynamic_teal_3():
    with pytest.raises(pt.TealInputError):
        pt.GeneratedID(pt.Int(0)).__teal__(avm3Options)


def test_gaid_dynamic():
    arg = pt.Int(0)
    expr = pt.GeneratedID(arg)
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [pt.TealOp(arg, pt.Op.int, 0), pt.TealOp(expr, pt.Op.gaids)]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_gaid_dynamic_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.GeneratedID(pt.Bytes("index"))



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *


def approval_program_explicit_ensure():
    args = [
        Bytes("base64", "iZWMx72KvU6Bw6sPAWQFL96YH+VMrBA0XKWD9XbZOZI="),
        Bytes(
            "base64",
            "if8ooA+32YZc4SQBvIDDY8tgTatPoq4IZ8Kr+We1t38LR2RuURmaVu9D4shbi4VvND87PUqq5/0vsNFEGIIEDA==",
        ),
        Addr("7JOPVEP3ABJUW5YZ5WFIONLPWTZ5MYX5HFK4K7JLGSIAG7RRB42MNLQ224"),
    ]
    opup = OpUp(OpUpMode.Explicit, Int(1))
    return Seq(
        If(Txn.application_id() != Int(0)).Then(
            Seq(
                opup.ensure_budget(Int(2000)),
                Pop(Ed25519Verify(args[0], args[1], args[2])),
            )
        ),
        Approve(),
    )


def approval_program_oncall_ensure():
    args = [
        Bytes("base64", "iZWMx72KvU6Bw6sPAWQFL96YH+VMrBA0XKWD9XbZOZI="),
        Bytes(
            "base64",
            "if8ooA+32YZc4SQBvIDDY8tgTatPoq4IZ8Kr+We1t38LR2RuURmaVu9D4shbi4VvND87PUqq5/0vsNFEGIIEDA==",
        ),
        Addr("7JOPVEP3ABJUW5YZ5WFIONLPWTZ5MYX5HFK4K7JLGSIAG7RRB42MNLQ224"),
    ]
    opup = OpUp(OpUpMode.OnCall)
    return Seq(
        If(Txn.application_id() != Int(0)).Then(
            Seq(
                opup.ensure_budget(Int(2000)),
                Pop(Ed25519Verify(args[0], args[1], args[2])),
            )
        ),
        Approve(),
    )


def approval_program_explicit_maximize():
    args = [
        Bytes("base64", "iZWMx72KvU6Bw6sPAWQFL96YH+VMrBA0XKWD9XbZOZI="),
        Bytes(
            "base64",
            "if8ooA+32YZc4SQBvIDDY8tgTatPoq4IZ8Kr+We1t38LR2RuURmaVu9D4shbi4VvND87PUqq5/0vsNFEGIIEDA==",
        ),
        Addr("7JOPVEP3ABJUW5YZ5WFIONLPWTZ5MYX5HFK4K7JLGSIAG7RRB42MNLQ224"),
    ]
    opup = OpUp(OpUpMode.Explicit, Int(1))
    return Seq(
        If(Txn.application_id() != Int(0)).Then(
            Seq(
                opup.maximize_budget(Int(3000)),
                Pop(Ed25519Verify(args[0], args[1], args[2])),
            )
        ),
        Approve(),
    )


def approval_program_oncall_maximize():
    args = [
        Bytes("base64", "iZWMx72KvU6Bw6sPAWQFL96YH+VMrBA0XKWD9XbZOZI="),
        Bytes(
            "base64",
            "if8ooA+32YZc4SQBvIDDY8tgTatPoq4IZ8Kr+We1t38LR2RuURmaVu9D4shbi4VvND87PUqq5/0vsNFEGIIEDA==",
        ),
        Addr("7JOPVEP3ABJUW5YZ5WFIONLPWTZ5MYX5HFK4K7JLGSIAG7RRB42MNLQ224"),
    ]
    opup = OpUp(OpUpMode.OnCall)
    return Seq(
        If(Txn.application_id() != Int(0)).Then(
            Seq(
                opup.maximize_budget(Int(3000)),
                Pop(Ed25519Verify(args[0], args[1], args[2])),
            )
        ),
        Approve(),
    )


if __name__ == "__main__":
    with open("program.teal", "w") as f:
        compiled = compileTeal(
            approval_program_oncall_maximize(), mode=Mode.Application, version=6
        )
        f.write(compiled)



#!/usr/bin/env python3

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyteal",
    version="0.19.0",
    author="Algorand",
    author_email="pypiservice@algorand.com",
    description="Algorand Smart Contracts in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/algorand/pyteal",
    packages=setuptools.find_packages(),
    install_requires=[
        # when changing this list, also update docs/requirements.txt
        "py-algorand-sdk>=1.9.0,<2.0.0",
        "semantic-version>=2.9.0,<3.0.0",
        "docstring-parser==0.14.1",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={"pyteal": ["*.pyi", "py.typed"]},
    python_requires=">=3.10",
)



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_tmpl_int():
    expr = pt.Tmpl.Int("TMPL_AMNT")
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.int, "TMPL_AMNT")])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_tmpl_int_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Tmpl.Int("whatever")


def test_tmpl_bytes():
    expr = pt.Tmpl.Bytes("TMPL_NOTE")
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "TMPL_NOTE")])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_tmpl_bytes_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Tmpl.Bytes("whatever")


def test_tmpl_addr():
    expr = pt.Tmpl.Addr("TMPL_RECEIVER0")
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.addr, "TMPL_RECEIVER0")])

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_tmpl_addr_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Tmpl.Addr("whatever")



import pytest

import pyteal as pt
from pyteal.errors import TealInputError

options = pt.CompileOptions()


def test_bytes_base32_no_padding():
    for s in (
        "ME",
        "MFRA",
        "MFRGG",
        "MFRGGZA",
        "MFRGGZDF",
        "7Z5PWO2C6LFNQFGHWKSK5H47IQP5OJW2M3HA2QPXTY3WTNP5NU2MHBW27M",
    ):
        expr = pt.Bytes("base32", s)
        assert expr.type_of() == pt.TealType.bytes
        expected = pt.TealSimpleBlock(
            [pt.TealOp(expr, pt.Op.byte, "base32(" + s + ")")]
        )
        actual, _ = expr.__teal__(options)
        assert actual == expected


def test_bytes_base32_padding():
    for s in (
        "ME======",
        "MFRA====",
        "MFRGG===",
        "MFRGGZA=",
        "7Z5PWO2C6LFNQFGHWKSK5H47IQP5OJW2M3HA2QPXTY3WTNP5NU2MHBW27M======",
    ):
        expr = pt.Bytes("base32", s)
        assert expr.type_of() == pt.TealType.bytes
        expected = pt.TealSimpleBlock(
            [pt.TealOp(expr, pt.Op.byte, "base32(" + s + ")")]
        )
        actual, _ = expr.__teal__(options)
        assert actual == expected


def test_bytes_base32_empty():
    expr = pt.Bytes("base32", "")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "base32()")])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_base64():
    expr = pt.Bytes("base64", "Zm9vYmE=")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "base64(Zm9vYmE=)")])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_base64_empty():
    expr = pt.Bytes("base64", "")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "base64()")])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_base16():
    expr = pt.Bytes("base16", "A21212EF")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "0xA21212EF")])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_base16_prefix():
    expr = pt.Bytes("base16", "0xA21212EF")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "0xA21212EF")])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_base16_empty():
    expr = pt.Bytes("base16", "")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "0x")])
    actual, _ = expr.__teal__(options)
    assert actual == expected


B16_ODD_LEN_TESTCASES = ["F", "0c1"]


@pytest.mark.parametrize("testcase", B16_ODD_LEN_TESTCASES)
def test_bytes_base16_odd_len(testcase):
    with pytest.raises(TealInputError):
        pt.Bytes("base16", testcase)


def test_bytes_utf8():
    expr = pt.Bytes("hello world")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, '"hello world"')])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_utf8_special_chars():
    expr = pt.Bytes("\t \n \r\n \\ \" ' 😀")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(
                expr, pt.Op.byte, '"\\t \\n \\r\\n \\\\ \\" \' \\xf0\\x9f\\x98\\x80"'
            )
        ]
    )
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_utf8_empty():
    expr = pt.Bytes("")
    assert expr.type_of() == pt.TealType.bytes
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, '""')])
    actual, _ = expr.__teal__(options)
    assert actual == expected


def test_bytes_raw():
    for value in (b"hello world", bytearray(b"hello world")):
        expr = pt.Bytes(value)
        assert expr.type_of() == pt.TealType.bytes
        expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "0x" + value.hex())])
        actual, _ = expr.__teal__(options)
        assert actual == expected


def test_bytes_raw_empty():
    for value in (b"", bytearray(b"")):
        expr = pt.Bytes(value)
        assert expr.type_of() == pt.TealType.bytes
        expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.byte, "0x")])
        actual, _ = expr.__teal__(options)
        assert actual == expected


def test_bytes_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Bytes("base16", b"FF")

    with pytest.raises(pt.TealInputError):
        pt.Bytes(b"base16", "FF")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base23", "")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "Zm9vYmE=")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "MFRGG====")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "MFRGG==")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "CCCCCC==")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "CCCCCC")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "C=======")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "C")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base32", "=")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base64", "?????")

    with pytest.raises(pt.TealInputError):
        pt.Bytes("base16", "7Z5PWO2C6LFNQFGHWKSK5H47IQP5OJW2M3HA2QPXTY3WTNP5NU2MHBW27M")



from enum import IntEnum
from typing import Union, Sequence, Literal, cast
from collections.abc import Sequence as CollectionSequence

from pyteal.errors import TealInputError

from pyteal.ast.assert_ import Assert
from pyteal.ast.bytes import Bytes
from pyteal.ast.int import Int
from pyteal.ast.seq import Seq
from pyteal.ast.unaryexpr import Len
from pyteal.ast.addr import Addr
from pyteal.ast.abi.type import ComputedValue, BaseType
from pyteal.ast.abi.array_static import StaticArray, StaticArrayTypeSpec
from pyteal.ast.abi.uint import ByteTypeSpec, Byte
from pyteal.ast.expr import Expr


class AddressLength(IntEnum):
    String = 58
    Bytes = 32


AddressLength.__module__ = "pyteal.abi"


class AddressTypeSpec(StaticArrayTypeSpec):
    def __init__(self) -> None:
        super().__init__(ByteTypeSpec(), AddressLength.Bytes)

    def new_instance(self) -> "Address":
        return Address()

    def annotation_type(self) -> "type[Address]":
        return Address

    def __str__(self) -> str:
        return "address"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, AddressTypeSpec)


AddressTypeSpec.__module__ = "pyteal.abi"


class Address(StaticArray[Byte, Literal[AddressLength.Bytes]]):
    def __init__(self) -> None:
        super().__init__(AddressTypeSpec())

    def type_spec(self) -> AddressTypeSpec:
        return AddressTypeSpec()

    def get(self) -> Expr:
        """Return the 32-byte value held by this Address as a PyTeal expression.

        The expression will have the type TealType.bytes.
        """
        return self.stored_value.load()

    def set(
        self,
        value: Union[
            str,
            bytes,
            Expr,
            Sequence[Byte],
            StaticArray[Byte, Literal[AddressLength.Bytes]],
            ComputedValue[StaticArray[Byte, Literal[AddressLength.Bytes]]],
            "Address",
            ComputedValue["Address"],
        ],
    ):
        """Set the value of this Address to the input value.

        The behavior of this method depends on the input argument type:

            * :code:`str`: set the value to the address from the encoded address string. This string must be a valid 58-character base-32 Algorand address with checksum.
            * :code:`bytes`: set the value to the raw address bytes. This byte string must have length 32.
            * :code:`Expr`: set the value to the result of a PyTeal expression, which must evaluate to a TealType.bytes. The program will fail if the evaluated byte string length is not 32.
            * :code:`Sequence[Byte]`: set the bytes of this Address to those contained in this Python sequence (e.g. a list or tuple). A compiler error will occur if the sequence length is not 32.
            * :code:`StaticArray[Byte, 32]`: copy the bytes from a StaticArray of 32 bytes.
            * :code:`ComputedValue[StaticArray[Byte, 32]]`: copy the bytes from a StaticArray of 32 bytes produced by a ComputedValue.
            * :code:`Address`: copy the value from another Address.
            * :code:`ComputedValue[Address]`: copy the value from an Address produced by a ComputedValue.

        Args:
            value: The new value this Address should take. This must follow the above constraints.

        Returns:
            An expression which stores the given value into this Address.
        """

        match value:
            case ComputedValue():
                pts = value.produced_type_spec()
                if pts == AddressTypeSpec() or pts == StaticArrayTypeSpec(
                    ByteTypeSpec(), AddressLength.Bytes
                ):
                    return value.store_into(self)

                raise TealInputError(
                    f"Got ComputedValue with type spec {pts}, expected AddressTypeSpec or StaticArray[Byte, Literal[AddressLength.Bytes]]"
                )
            case BaseType():
                if (
                    value.type_spec() == AddressTypeSpec()
                    or value.type_spec()
                    == StaticArrayTypeSpec(ByteTypeSpec(), AddressLength.Bytes)
                ):
                    return self.stored_value.store(value.stored_value.load())

                raise TealInputError(
                    f"Got {value} with type spec {value.type_spec()}, expected AddressTypeSpec"
                )
            case str():
                # Addr throws if value is invalid address
                return self.stored_value.store(Addr(value))
            case bytes():
                if len(value) == AddressLength.Bytes:
                    return self.stored_value.store(Bytes(value))
                raise TealInputError(
                    f"Got bytes with length {len(value)}, expected {AddressLength.Bytes}"
                )
            case Expr():
                return Seq(
                    self.stored_value.store(value),
                    Assert(
                        Len(self.stored_value.load()) == Int(AddressLength.Bytes.value)
                    ),
                )
            case CollectionSequence():
                if len(value) != AddressLength.Bytes:
                    raise TealInputError(
                        f"Got bytes with length {len(value)}, expected {AddressLength.Bytes}"
                    )
                return super().set(cast(Sequence[Byte], value))

        raise TealInputError(
            f"Got {type(value)}, expected Sequence, StaticArray, ComputedValue, Address, str, bytes, Expr"
        )


Address.__module__ = "pyteal.abi"



from abc import ABC, abstractmethod

from pyteal.ast.expr import Expr


class Array(ABC):
    """Represents a variable length array of objects."""

    @abstractmethod
    def length(self) -> Expr:
        """Get the length of the array."""
        pass

    @abstractmethod
    def __getitem__(self, index: int):
        """Get the value at a given index in this array."""
        pass


Array.__module__ = "pyteal"



from typing import Union, Tuple, cast, TYPE_CHECKING

from pyteal.types import TealType, require_type
from pyteal.errors import verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class BinaryExpr(Expr):
    """An expression with two arguments."""

    def __init__(
        self,
        op: Op,
        inputType: Union[TealType, Tuple[TealType, TealType]],
        outputType: TealType,
        argLeft: Expr,
        argRight: Expr,
    ) -> None:
        super().__init__()
        if type(inputType) is tuple:
            leftType, rightType = inputType
        else:
            leftType = cast(TealType, inputType)
            rightType = leftType
        require_type(argLeft, leftType)
        require_type(argRight, rightType)

        self.op = op
        self.outputType = outputType
        self.argLeft = argLeft
        self.argRight = argRight

    def __teal__(self, options: "CompileOptions"):
        verifyProgramVersion(
            self.op.min_version,
            options.version,
            "Program version too low to use op {}".format(self.op),
        )

        return TealBlock.FromOp(
            options, TealOp(self, self.op), self.argLeft, self.argRight
        )

    def __str__(self):
        return "({} {} {})".format(
            str(self.op).title().replace("_", ""), self.argLeft, self.argRight
        )

    def type_of(self):
        return self.outputType

    def has_return(self):
        return False


BinaryExpr.__module__ = "pyteal"


def Minus(left: Expr, right: Expr) -> Expr:
    """Subtract two numbers.

    Produces left - right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.minus, TealType.uint64, TealType.uint64, left, right)


def Div(left: Expr, right: Expr) -> Expr:
    """Divide two numbers.

    Produces left / right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.div, TealType.uint64, TealType.uint64, left, right)


def Mod(left: Expr, right: Expr) -> Expr:
    """Modulo expression.

    Produces left % right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.mod, TealType.uint64, TealType.uint64, left, right)


def Exp(a: Expr, b: Expr) -> Expr:
    """Exponential expression.

    Produces a ** b.

    Requires program version 4 or higher.

    Args:
        a: Must evaluate to uint64.
        b: Must evaluate to uint64.
    """
    return BinaryExpr(Op.exp, TealType.uint64, TealType.uint64, a, b)


def BitwiseAnd(left: Expr, right: Expr) -> Expr:
    """Bitwise and expression.

    Produces left & right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.bitwise_and, TealType.uint64, TealType.uint64, left, right)


def BitwiseOr(left: Expr, right: Expr) -> Expr:
    """Bitwise or expression.

    Produces left | right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.bitwise_or, TealType.uint64, TealType.uint64, left, right)


def BitwiseXor(left: Expr, right: Expr) -> Expr:
    """Bitwise xor expression.

    Produces left ^ right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.bitwise_xor, TealType.uint64, TealType.uint64, left, right)


def ShiftLeft(a: Expr, b: Expr) -> Expr:
    """Bitwise left shift expression.

    Produces a << b. This is equivalent to a times 2^b, modulo 2^64.

    Requires program version 4 or higher.

    Args:
        a: Must evaluate to uint64.
        b: Must evaluate to uint64.
    """
    return BinaryExpr(Op.shl, TealType.uint64, TealType.uint64, a, b)


def ShiftRight(a: Expr, b: Expr) -> Expr:
    """Bitwise right shift expression.

    Produces a >> b. This is equivalent to a divided by 2^b.

    Requires program version 4 or higher.

    Args:
        a: Must evaluate to uint64.
        b: Must evaluate to uint64.
    """
    return BinaryExpr(Op.shr, TealType.uint64, TealType.uint64, a, b)


def Eq(left: Expr, right: Expr) -> Expr:
    """Equality expression.

    Checks if left == right.

    Args:
        left: A value to check.
        right: The other value to check. Must evaluate to the same type as left.
    """
    return BinaryExpr(Op.eq, right.type_of(), TealType.uint64, left, right)


def Neq(left: Expr, right: Expr) -> Expr:
    """Difference expression.

    Checks if left != right.

    Args:
        left: A value to check.
        right: The other value to check. Must evaluate to the same type as left.
    """
    return BinaryExpr(Op.neq, right.type_of(), TealType.uint64, left, right)


def Lt(left: Expr, right: Expr) -> Expr:
    """Less than expression.

    Checks if left < right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.lt, TealType.uint64, TealType.uint64, left, right)


def Le(left: Expr, right: Expr) -> Expr:
    """Less than or equal to expression.

    Checks if left <= right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.le, TealType.uint64, TealType.uint64, left, right)


def Gt(left: Expr, right: Expr) -> Expr:
    """Greater than expression.

    Checks if left > right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.gt, TealType.uint64, TealType.uint64, left, right)


def Ge(left: Expr, right: Expr) -> Expr:
    """Greater than or equal to expression.

    Checks if left >= right.

    Args:
        left: Must evaluate to uint64.
        right: Must evaluate to uint64.
    """
    return BinaryExpr(Op.ge, TealType.uint64, TealType.uint64, left, right)


def GetBit(value: Expr, index: Expr) -> Expr:
    """Get the bit value of an expression at a specific index.

    The meaning of index differs if value is an integer or a byte string.

    * For integers, bit indexing begins with low-order bits. For example, :code:`GetBit(Int(16), Int(4))`
      yields 1. Any other valid index would yield a bit value of 0. Any integer less than 64 is a
      valid index.

    * For byte strings, bit indexing begins at the first byte. For example, :code:`GetBit(Bytes("base16", "0xf0"), Int(0))`
      yields 1. Any index less than 4 would yield 1, and any valid index 4 or greater would yield 0.
      Any integer less than 8*Len(value) is a valid index.

    Requires program version 3 or higher.

    Args:
        value: The value containing bits. Can evaluate to any type.
        index: The index of the bit to extract. Must evaluate to uint64.
    """
    return BinaryExpr(
        Op.getbit, (TealType.anytype, TealType.uint64), TealType.uint64, value, index
    )


def GetByte(value: Expr, index: Expr) -> Expr:
    """Extract a single byte as an integer from a byte string.

    Similar to GetBit, indexing begins at the first byte. For example, :code:`GetByte(Bytes("base16", "0xff0000"), Int(0))`
    yields 255. Any other valid index would yield 0.

    Requires program version 3 or higher.

    Args:
        value: The value containing the bytes. Must evaluate to bytes.
        index: The index of the byte to extract. Must evaluate to an integer less than Len(value).
    """
    return BinaryExpr(
        Op.getbyte, (TealType.bytes, TealType.uint64), TealType.uint64, value, index
    )


def BytesAdd(left: Expr, right: Expr) -> Expr:
    """Add two numbers as bytes.

    Produces left + right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_add, TealType.bytes, TealType.bytes, left, right)


def BytesMinus(left: Expr, right: Expr) -> Expr:
    """Subtract two numbers as bytes.

    Produces left - right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_minus, TealType.bytes, TealType.bytes, left, right)


def BytesDiv(left: Expr, right: Expr) -> Expr:
    """Divide two numbers as bytes.

    Produces left / right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Panics if right is 0.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_div, TealType.bytes, TealType.bytes, left, right)


def BytesMul(left: Expr, right: Expr) -> Expr:
    """Multiply two numbers as bytes.

    Produces left * right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_mul, TealType.bytes, TealType.bytes, left, right)


def BytesMod(left: Expr, right: Expr) -> Expr:
    """Modulo expression with bytes as arguments.

    Produces left % right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Panics if right is 0.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_mod, TealType.bytes, TealType.bytes, left, right)


def BytesAnd(left: Expr, right: Expr) -> Expr:
    """Bitwise and expression with bytes as arguments.

    Produces left & right.
    Left and right are zero-left extended to the greater of their lengths.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_and, TealType.bytes, TealType.bytes, left, right)


def BytesOr(left: Expr, right: Expr) -> Expr:
    """Bitwise or expression with bytes as arguments.

    Produces left | right.
    Left and right are zero-left extended to the greater of their lengths.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_or, TealType.bytes, TealType.bytes, left, right)


def BytesXor(left: Expr, right: Expr) -> Expr:
    """Bitwise xor expression with bytes as arguments.

    Produces left ^ right.
    Left and right are zero-left extended to the greater of their lengths.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_xor, TealType.bytes, TealType.bytes, left, right)


def BytesEq(left: Expr, right: Expr) -> Expr:
    """Equality expression with bytes as arguments.

    Checks if left == right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_eq, TealType.bytes, TealType.uint64, left, right)


def BytesNeq(left: Expr, right: Expr) -> Expr:
    """Difference expression with bytes as arguments.

    Checks if left != right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_neq, TealType.bytes, TealType.uint64, left, right)


def BytesLt(left: Expr, right: Expr) -> Expr:
    """Less than expression with bytes as arguments.

    Checks if left < right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_lt, TealType.bytes, TealType.uint64, left, right)


def BytesLe(left: Expr, right: Expr) -> Expr:
    """Less than or equal to expression with bytes as arguments.

    Checks if left <= right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_le, TealType.bytes, TealType.uint64, left, right)


def BytesGt(left: Expr, right: Expr) -> Expr:
    """Greater than expression with bytes as arguments.

    Checks if left > right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_gt, TealType.bytes, TealType.uint64, left, right)


def BytesGe(left: Expr, right: Expr) -> Expr:
    """Greater than or equal to expression with bytes as arguments.

    Checks if left >= right, where left and right are interpreted as big-endian unsigned integers.
    Arguments must not exceed 64 bytes.

    Requires program version 4 or higher.

    Args:
        left: Must evaluate to bytes.
        right: Must evaluate to bytes.
    """
    return BinaryExpr(Op.b_ge, TealType.bytes, TealType.uint64, left, right)


def ExtractUint16(string: Expr, offset: Expr) -> Expr:
    """Extract 2 bytes (16 bits) and convert them to an integer.

    The bytes starting at :code:`offset` up to but not including :code:`offset + 2` will be
    interpreted as a big-endian unsigned integer.

    If :code:`offset + 2` exceeds :code:`Len(string)`, the program fails.

    Requires program version 5 or higher.

    Args:
        string: The bytestring to extract from. Must evaluate to bytes.
        offset: The offset in the bytestring to start extracing. Must evaluate to uint64.
    """
    return BinaryExpr(
        Op.extract_uint16,
        (TealType.bytes, TealType.uint64),
        TealType.uint64,
        string,
        offset,
    )


def ExtractUint32(string: Expr, offset: Expr) -> Expr:
    """Extract 4 bytes (32 bits) and convert them to an integer.

    The bytes starting at :code:`offset` up to but not including :code:`offset + 4` will be
    interpreted as a big-endian unsigned integer.

    If :code:`offset + 4` exceeds :code:`Len(string)`, the program fails.

    Requires program version 5 or higher.

    Args:
        string: The bytestring to extract from. Must evaluate to bytes.
        offset: The offset in the bytestring to start extracing. Must evaluate to uint64.
    """
    return BinaryExpr(
        Op.extract_uint32,
        (TealType.bytes, TealType.uint64),
        TealType.uint64,
        string,
        offset,
    )


def ExtractUint64(string: Expr, offset: Expr) -> Expr:
    """Extract 8 bytes (64 bits) and convert them to an integer.

    The bytes starting at :code:`offset` up to but not including :code:`offset + 8` will be
    interpreted as a big-endian unsigned integer.

    If :code:`offset + 8` exceeds :code:`Len(string)`, the program fails.

    Requires program version 5 or higher.

    Args:
        string: The bytestring to extract from. Must evaluate to bytes.
        offset: The offset in the bytestring to start extracing. Must evaluate to uint64.
    """
    return BinaryExpr(
        Op.extract_uint64,
        (TealType.bytes, TealType.uint64),
        TealType.uint64,
        string,
        offset,
    )



from abc import ABC, abstractmethod
from typing import List, Optional, TYPE_CHECKING
from contextlib import AbstractContextManager

if TYPE_CHECKING:
    from pyteal.ast import Expr, ScratchSlot, SubroutineDefinition


class TealComponent(ABC):
    def __init__(self, expr: Optional["Expr"]):
        self.expr = expr

    def getSlots(self) -> List["ScratchSlot"]:
        return []

    def assignSlot(self, slot: "ScratchSlot", location: int) -> None:
        pass

    def getSubroutines(self) -> List["SubroutineDefinition"]:
        return []

    def resolveSubroutine(self, subroutine: "SubroutineDefinition", label: str) -> None:
        pass

    @abstractmethod
    def assemble(self) -> str:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass

    class Context:

        checkExprEquality = True

        class ExprEqualityContext(AbstractContextManager):
            def __enter__(self):
                TealComponent.Context.checkExprEquality = False
                return self

            def __exit__(self, *args):
                TealComponent.Context.checkExprEquality = True
                return None

        @classmethod
        def ignoreExprEquality(cls):
            return cls.ExprEqualityContext()

        checkScratchSlotEquality = True

        class ScratchSlotEqualityContext(AbstractContextManager):
            def __enter__(self):
                TealComponent.Context.checkScratchSlotEquality = False

            def __exit__(self, *args):
                TealComponent.Context.checkScratchSlotEquality = True
                return None

        @classmethod
        def ignoreScratchSlotEquality(cls):
            """When comparing TealOps, do not verify the equality of any ScratchSlot arguments.

            This is commonly used in testing to verify the that two control flow graphs contains the
            same operations, but may use different ScratchSlots in them. In this case, you will most
            likely want to also use use the following code after comparing with this option enabled:

                .. code-block:: python

                    TealBlock.MatchScratchSlotReferences(
                        TealBlock.GetReferencedScratchSlots(actual),
                        TealBlock.GetReferencedScratchSlots(expected),
                    )

            This ensures that the ScratchSlot usages between the two control flow graphs is
            equivalent. See :any:`TealBlock.MatchScratchSlotReferences` for more info.
            """
            return cls.ScratchSlotEqualityContext()


TealComponent.__module__ = "pyteal"



import pytest

import pyteal as pt
from pyteal.ast.seq import _use_seq_if_multiple

options = pt.CompileOptions()


def test_seq_zero():
    for expr in (pt.Seq(), pt.Seq([])):
        assert expr.type_of() == pt.TealType.none
        assert not expr.has_return()

        expected = pt.TealSimpleBlock([])

        actual, _ = expr.__teal__(options)

        assert actual == expected


def test_seq_one():
    items = [pt.Int(0)]
    expr = pt.Seq(items)
    assert expr.type_of() == pt.TealType.uint64

    expected, _ = items[0].__teal__(options)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_seq_two():
    items = [pt.App.localPut(pt.Int(0), pt.Bytes("key"), pt.Int(1)), pt.Int(7)]
    expr = pt.Seq(items)
    assert expr.type_of() == items[-1].type_of()

    expected, first_end = items[0].__teal__(options)
    first_end.setNextBlock(items[1].__teal__(options)[0])
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_seq_three():
    items = [
        pt.App.localPut(pt.Int(0), pt.Bytes("key1"), pt.Int(1)),
        pt.App.localPut(pt.Int(1), pt.Bytes("key2"), pt.Bytes("value2")),
        pt.Pop(pt.Bytes("end")),
    ]
    expr = pt.Seq(items)
    assert expr.type_of() == items[-1].type_of()

    expected, first_end = items[0].__teal__(options)
    second_start, second_end = items[1].__teal__(options)
    first_end.setNextBlock(second_start)
    third_start, _ = items[2].__teal__(options)
    second_end.setNextBlock(third_start)

    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_seq_has_return():
    exprWithReturn = pt.Seq(
        [
            pt.App.localPut(pt.Int(0), pt.Bytes("key1"), pt.Int(1)),
            pt.Return(pt.Int(1)),
        ]
    )
    assert exprWithReturn.has_return()

    exprWithoutReturn = pt.Seq(
        [pt.App.localPut(pt.Int(0), pt.Bytes("key1"), pt.Int(1)), pt.Int(1)]
    )
    assert not exprWithoutReturn.has_return()


def test_seq_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Seq([pt.Int(1), pt.Pop(pt.Int(2))])

    with pytest.raises(pt.TealTypeError):
        pt.Seq([pt.Int(1), pt.Int(2)])

    with pytest.raises(pt.TealTypeError):
        pt.Seq([pt.Seq([pt.Pop(pt.Int(1)), pt.Int(2)]), pt.Int(3)])


def test_seq_overloads_equivalence():
    items = [
        pt.App.localPut(pt.Int(0), pt.Bytes("key1"), pt.Int(1)),
        pt.App.localPut(pt.Int(1), pt.Bytes("key2"), pt.Bytes("value2")),
        pt.Pop(pt.Bytes("end")),
    ]
    expr1 = pt.Seq(items)
    expr2 = pt.Seq(*items)

    expected = expr1.__teal__(options)
    actual = expr2.__teal__(options)

    assert actual == expected


def test_use_seq_if_multiple():
    # Test a single expression
    items = [pt.Int(1)]
    expr = _use_seq_if_multiple(*items)

    expected = items[0].__teal__(options)
    actual = expr.__teal__(options)

    assert actual == expected

    # Test multiple expressions
    items = [pt.Pop(pt.Int(1)), pt.Int(2)]
    expr = _use_seq_if_multiple(*items)

    expected = pt.Seq(items).__teal__(options)
    actual = expr.__teal__(options)

    assert actual == expected


def test_use_seq_if_multiple_overloads_equivalence():
    items = [pt.Pop(pt.Int(1)), pt.Int(2)]
    expr = _use_seq_if_multiple(items)

    expected = pt.Seq(items).__teal__(options)
    actual = expr.__teal__(options)

    assert actual == expected



from pyteal.errors import TealInputError
from pyteal.types import TealType, require_type

from pyteal.ast.expr import Expr
from pyteal.ast.scratch import ScratchSlot, ScratchLoad, ScratchStore


class ScratchVar:
    """
    Interface around Scratch space, similar to get/put local/global state

    Example:
        .. code-block:: python

            myvar = ScratchVar(TealType.uint64)
            Seq([
                myvar.store(Int(5)),
                Assert(myvar.load() == Int(5))
            ])
    """

    def __init__(self, type: TealType = TealType.anytype, slotId: int = None):
        """Create a new ScratchVar with an optional type.

        Args:
            type (optional): The type that this variable can hold. An error will be thrown if an
                expression with an incompatible type is stored in this variable. Defaults to
                TealType.anytype.
            slotId (optional): A scratch slot id that the compiler must store the value.
                This id may be a Python int in the range [0-256).
        """
        self.slot = ScratchSlot(requestedSlotId=slotId)
        self.type = type

    def storage_type(self) -> TealType:
        """Get the type of expressions that can be stored in this ScratchVar."""
        return self.type

    def store(self, value: Expr) -> Expr:
        """Store value in Scratch Space

        Args:
            value: The value to store. Must conform to this ScratchVar's type.
        """
        require_type(value, self.type)
        return self.slot.store(value)

    def load(self) -> ScratchLoad:
        """Load value from Scratch Space"""
        return self.slot.load(self.type)

    def index(self) -> Expr:
        return self.slot.index()


ScratchVar.__module__ = "pyteal"


class DynamicScratchVar(ScratchVar):
    """
    Example of Dynamic Scratch space whereby the slot index is picked up from the stack:
        .. code-block:: python

            player_score = DynamicScratchVar(TealType.uint64)

            wilt = ScratchVar(TealType.uint64, 129)
            kobe = ScratchVar(TealType.uint64)
            dt = ScratchVar(TealType.uint64, 131)

            seq = Seq(
                player_score.set_index(wilt),
                player_score.store(Int(100)),
                player_score.set_index(kobe),
                player_score.store(Int(81)),
                player_score.set_index(dt),
                player_score.store(Int(73)),
                Assert(player_score.load() == Int(73)),
                Assert(player_score.index() == Int(131)),
                player_score.set_index(wilt),
                Assert(player_score.load() == Int(100)),
                Assert(player_score.index() == Int(129)),
                Int(100),
            )
    """

    def __init__(self, ttype: TealType = TealType.anytype):
        """Create a new DynamicScratchVar which references other ScratchVar's

        Args:
            ttype (optional): The type that this variable can hold. Defaults to TealType.anytype.
        """
        super().__init__(TealType.uint64)
        self.dynamic_type = ttype  # differentiates from ScratchVar.type

    def set_index(self, index_var: ScratchVar) -> Expr:
        """Set this DynamicScratchVar to reference the provided `index_var`.
        Followup `store`, `load` and `index` operations will use the provided `index_var` until
        `set_index()` is called again to reset the referenced ScratchVar.
        """
        # Explanatory comment per Issue #242: Preliminary evidence shows that letting users
        # pass in any ScratchVar subtype (i.e. DynamicScratchVar) may in fact work.
        # However, we are leaving this guard in place pending further investigation.
        # TODO: gain confidence that DynamicScratchVar can be used here and
        # modify the below strict type equality to `isinstance(index_var, ScratchVar)`
        if type(index_var) is not ScratchVar:
            raise TealInputError(
                "Only allowed to use ScratchVar objects for setting indices, but was given a {}".format(
                    type(index_var)
                )
            )

        return super().store(index_var.index())

    def storage_type(self) -> TealType:
        """Get the type of expressions that can be stored in this ScratchVar."""
        return self.dynamic_type

    def store(self, value: Expr) -> Expr:
        """Store the value in the referenced ScratchVar."""
        require_type(value, self.dynamic_type)
        return ScratchStore(slot=None, value=value, index_expression=self.index())

    def load(self) -> ScratchLoad:
        """Load the current value from the referenced ScratchVar."""
        return ScratchLoad(
            slot=None, type=self.dynamic_type, index_expression=self.index()
        )

    def index(self) -> Expr:
        """Get the index of the referenced ScratchVar."""
        return super().load()

    def internal_index(self) -> Expr:
        """Get the index of _this_ DynamicScratchVar, as opposed to that of the referenced ScratchVar."""
        return super().index()


DynamicScratchVar.__module__ = "pyteal"



import random

from graviton.blackbox import DryRunInspector

import pyteal as pt
from pyteal.ast.subroutine import ABIReturnSubroutine

from tests.blackbox import (
    Blackbox,
    PyTealDryRunExecutor,
)


# ---- Integers and Complex Integral Numbers (aka Gaussian Integers) ---- #


"""
WARNING: The following ABI types Int65 and Complex130 are ONLY for the purpose of testing/demo'ing 
ABISubroutine and graviton ABI capabilities and are NOT the recommended approach for 
implementing integers and complex integers.
A better approach would likely leverage `Uint64` (if any ABI type at all) and make use of 2's complement arithmetic.

Cf. https://github.com/algorand/pyteal/issues/184
"""

Int65 = pt.abi.Tuple2[pt.abi.Bool, pt.abi.Uint64]
Complex130 = pt.abi.Tuple2[Int65, Int65]


@Blackbox(input_types=[None, None])
@pt.ABIReturnSubroutine
def int65_minus_cond(x: Int65, y: Int65, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    x0 = pt.abi.Bool()
    x1 = pt.abi.Uint64()
    y0 = pt.abi.Bool()
    y1 = pt.abi.Uint64()
    z0 = pt.abi.Bool()
    z1 = pt.abi.Uint64()
    return pt.Seq(
        x0.set(x[0]),
        x1.set(x[1]),
        y0.set(y[0]),
        y1.set(y[1]),
        pt.Cond(
            # Case I. x, y positive
            [
                pt.And(x0.get(), y0.get()),
                pt.Seq(
                    z0.set(x1.get() >= y1.get()),
                    z1.set(
                        pt.If(x1.get() <= y1.get())
                        .Then(y1.get() - x1.get())
                        .Else(x1.get() - y1.get())
                    ),
                ),
            ],
            # Case II. x positive, y negative
            [
                pt.And(x0.get(), pt.Not(y0.get())),
                pt.Seq(z0.set(True), z1.set(x1.get() + y1.get())),
            ],
            # Case III. x negative, y positive
            [
                pt.And(pt.Not(x0.get()), y0.get()),
                pt.Seq(z0.set(False), z1.set(x1.get() + y1.get())),
            ],
            # Case IV. x, y negative
            [
                pt.Int(1),
                pt.Seq(
                    z0.set(x1.get() <= y1.get()),
                    z1.set(
                        pt.If(x1.get() <= y1.get())
                        .Then(y1.get() - x1.get())
                        .Else(x1.get() - y1.get())
                    ),
                ),
            ],
        ),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None, None])
@pt.ABIReturnSubroutine
def int65_sub(x: Int65, y: Int65, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    x0 = pt.abi.Bool()
    x1 = pt.abi.Uint64()
    y0 = pt.abi.Bool()
    y1 = pt.abi.Uint64()
    z0 = pt.abi.Bool()
    z1 = pt.abi.Uint64()
    return pt.Seq(
        x0.set(x[0]),
        x1.set(x[1]),
        y0.set(y[0]),
        y1.set(y[1]),
        pt.If(x0.get() == y0.get())
        .Then(  # Case I. x, y same signature
            pt.Seq(
                z0.set(pt.Not(x0.get()) ^ (x1.get() >= y1.get())),
                z1.set(
                    pt.If(x1.get() <= y1.get())
                    .Then(y1.get() - x1.get())
                    .Else(x1.get() - y1.get())
                ),
            )
        )
        .Else(  # Case II. x, y opposite signatures
            pt.Seq(
                z0.set(x0.get()),
                z1.set(x1.get() + y1.get()),
            ),
        ),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None, None])
@pt.ABIReturnSubroutine
def int65_mult(x: Int65, y: Int65, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    # TODO: can we get something like the following one-liner working?
    # return output.set(pt.Not(x[0].get() ^ y[0].get()), x[1].get() * y[1].get())
    def get(x):
        return x.use(lambda ni: ni.get())

    return pt.Seq(
        (z0 := pt.abi.Bool()).set(pt.Not(get(x[0]) ^ get(y[0]))),
        (z1 := pt.abi.Uint64()).set(get(x[1]) * get(y[1])),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None])
@ABIReturnSubroutine
def int65_negate(x: Int65, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    # TODO: can I haz a one-liner pls????
    x0 = pt.abi.Bool()
    x1 = pt.abi.Uint64()
    z0 = pt.abi.Bool()
    z1 = pt.abi.Uint64()
    return pt.Seq(
        x0.set(x[0]),
        x1.set(x[1]),
        z0.set(pt.Not(x0.get())),
        z1.set(x1.get()),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None, None])
@ABIReturnSubroutine
def int65_add(x: Int65, y: Int65, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    return pt.Seq(y.set(int65_negate(y)), output.set(int65_sub(x, y)))


@Blackbox(input_types=[None, None])
@ABIReturnSubroutine
def complex130_add(x: Complex130, y: Complex130, *, output: Complex130):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    x0 = pt.abi.make(Int65)
    x1 = pt.abi.make(Int65)
    y0 = pt.abi.make(Int65)
    y1 = pt.abi.make(Int65)
    z0 = pt.abi.make(Int65)
    z1 = pt.abi.make(Int65)
    return pt.Seq(
        x0.set(x[0]),
        x1.set(x[1]),
        y0.set(y[0]),
        y1.set(y[1]),
        z0.set(int65_add(x0, y0)),
        z1.set(int65_add(x1, y1)),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None, None])
@ABIReturnSubroutine
def complex130_mult(x: Complex130, y: Complex130, *, output: Complex130):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    x0 = pt.abi.make(Int65)
    x1 = pt.abi.make(Int65)
    y0 = pt.abi.make(Int65)
    y1 = pt.abi.make(Int65)
    t1 = pt.abi.make(Int65)
    t2 = pt.abi.make(Int65)
    t3 = pt.abi.make(Int65)
    t4 = pt.abi.make(Int65)
    z0 = pt.abi.make(Int65)
    z1 = pt.abi.make(Int65)
    return pt.Seq(
        x0.set(x[0]),
        x1.set(x[1]),
        y0.set(y[0]),
        y1.set(y[1]),
        # TODO: why can't I chain ABI calls?
        # z0.set(int65_sub(int65_mult(x0, y0), int65_mult(x1, y1))),
        # z1.set(int65_add(int65_mult(x0, y1), int65_mult(x1, y0))),
        t1.set(int65_mult(x0, y0)),
        t2.set(int65_mult(x1, y1)),
        t3.set(int65_mult(x0, y1)),
        t4.set(int65_mult(x1, y0)),
        z0.set(int65_sub(t1, t2)),
        z1.set(int65_add(t3, t4)),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None])
@ABIReturnSubroutine
def complex130_real(x: Complex130, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    return output.set(x[0])


@Blackbox(input_types=[None])
@ABIReturnSubroutine
def complex130_imag(x: Complex130, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    return output.set(x[1])


@Blackbox(input_types=[None])
@ABIReturnSubroutine
def complex130_conjugate(x: Complex130, *, output: Complex130):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    z0 = pt.abi.make(Int65)
    z1 = pt.abi.make(Int65)
    return pt.Seq(
        z0.set(complex130_real(x)),
        z1.set(complex130_imag(x)),
        z1.set(int65_negate(z1)),
        output.set(z0, z1),
    )


@Blackbox(input_types=[None])
@ABIReturnSubroutine
def complex130_norm_squared(x: Complex130, *, output: Int65):
    """
    WARNING: not an ideal implementation. See explanation in first WARNING above.
    """
    t = pt.abi.make(Complex130)
    return pt.Seq(
        t.set(complex130_conjugate(x)),
        t.set(complex130_mult(t, x)),
        output.set(complex130_real(t)),
    )


# ---- additional stand-alone ABIReturnSubroutine's ---- #


@Blackbox(input_types=[None])
@ABIReturnSubroutine
def conditional_factorial(_factor: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    i = pt.ScratchVar(pt.TealType.uint64)

    return pt.Seq(
        output.set(1),
        pt.If(_factor.get() <= pt.Int(1))
        .Then(pt.Return())
        .Else(
            pt.For(
                i.store(_factor.get()),
                i.load() > pt.Int(1),
                i.store(i.load() - pt.Int(1)),
            ).Do(output.set(output.get() * i.load())),
        ),
    )


# ---- integration test functions ---- #


def test_integer65():
    bbpt_subtract_slick = PyTealDryRunExecutor(int65_sub, pt.Mode.Application)

    bbpt_subtract_cond = PyTealDryRunExecutor(int65_minus_cond, pt.Mode.Application)

    bbpt_mult = PyTealDryRunExecutor(int65_mult, pt.Mode.Application)

    bbpt_negate = PyTealDryRunExecutor(int65_negate, pt.Mode.Application)

    bbpt_add = PyTealDryRunExecutor(int65_add, pt.Mode.Application)

    def pynum_to_tuple(n):
        return (n >= 0, abs(n))

    def pytuple_to_num(t):
        s, x = t
        return x if s else -x

    N = 100
    random.seed(42)

    choices = range(-9_999, 10_000)
    unary_inputs = [(pynum_to_tuple(x),) for x in random.sample(choices, N)]

    binary_inputs = [
        (pynum_to_tuple(x), pynum_to_tuple(y))
        for x, y in zip(random.sample(choices, N), random.sample(choices, N))
    ]

    def binary_dryrun(p: PyTealDryRunExecutor) -> list[DryRunInspector]:
        return p.dryrun_on_sequence(binary_inputs)

    # Binary:
    inspectors_subtract_slick = binary_dryrun(bbpt_subtract_slick)

    inspectors_subtract_cond = binary_dryrun(bbpt_subtract_cond)

    inspectors_mult = binary_dryrun(bbpt_mult)

    inspectors_add = binary_dryrun(bbpt_add)

    # Unary:
    inspectors_negate = bbpt_negate.dryrun_on_sequence(unary_inputs)

    for i in range(N):
        binary_args = binary_inputs[i]
        x, y = tuple(map(pytuple_to_num, binary_args))

        unary_args = unary_inputs[i]
        u = pytuple_to_num(unary_args[0])

        inspector_subtract_slick = inspectors_subtract_slick[i]
        inspector_subtract_cond = inspectors_subtract_cond[i]
        inspector_mult = inspectors_mult[i]
        inspector_add = inspectors_add[i]

        inspector_negate = inspectors_negate[i]

        assert x - y == pytuple_to_num(
            inspector_subtract_slick.last_log()
        ), inspector_subtract_slick.report(
            binary_args, f"failed for {binary_args}", row=i
        )

        assert x - y == pytuple_to_num(
            inspector_subtract_cond.last_log()
        ), inspector_subtract_cond.report(
            binary_args, f"failed for {binary_args}", row=i
        )

        assert x * y == pytuple_to_num(
            inspector_mult.last_log()
        ), inspector_mult.report(binary_args, f"failed for {binary_args}", row=i)

        assert x + y == pytuple_to_num(inspector_add.last_log()), inspector_add.report(
            binary_args, f"failed for {binary_args}", row=i
        )

        assert -u == pytuple_to_num(
            inspector_negate.last_log()
        ), inspector_negate.report(unary_args, f"failed for {unary_args}", row=i)


def test_complex130():
    # Binary:

    bbpt_cplx_add = PyTealDryRunExecutor(complex130_add, pt.Mode.Application)

    bbpt_cplx_mult = PyTealDryRunExecutor(complex130_mult, pt.Mode.Application)

    # Unary:

    bbpt_complex_real = PyTealDryRunExecutor(complex130_real, pt.Mode.Application)

    bbpt_complex_imag = PyTealDryRunExecutor(complex130_imag, pt.Mode.Application)

    bbpt_complex_conjugate = PyTealDryRunExecutor(
        complex130_conjugate, pt.Mode.Application
    )

    bbpt_complex_norm_squared = PyTealDryRunExecutor(
        complex130_norm_squared, pt.Mode.Application
    )

    def pyint_to_tuple(n):
        return (n >= 0, abs(n))

    def pycomplex_to_tuple(z):
        return (pyint_to_tuple(int(z.real)), pyint_to_tuple(int(z.imag)))

    def pytuple_to_int(t):
        s, x = t
        return x if s else -x

    def pytuple_to_complex(tt):
        tx, ty = tt
        return complex(pytuple_to_int(tx), pytuple_to_int(ty))

    N = 100
    # just for fun - no random seed - but this shouldn't be flakey

    choices = range(-999_999, 1_000_000)

    unary_inputs = [
        (pycomplex_to_tuple(complex(x, y)),)
        for x, y in zip(random.sample(choices, N), random.sample(choices, N))
    ]

    binary_inputs = [
        (pycomplex_to_tuple(complex(x, y)), pycomplex_to_tuple(complex(z, w)))
        for x, y, z, w in zip(
            random.sample(choices, N),
            random.sample(choices, N),
            random.sample(choices, N),
            random.sample(choices, N),
        )
    ]

    # Binary:
    def binary_dryrun(p: PyTealDryRunExecutor) -> list[DryRunInspector]:
        return p.dryrun_on_sequence(binary_inputs)

    inspectors_cplx_add = binary_dryrun(bbpt_cplx_add)

    inspectors_cplx_mult = binary_dryrun(bbpt_cplx_mult)

    # Unary:
    def unary_dryrun(p: PyTealDryRunExecutor) -> list[DryRunInspector]:
        return p.dryrun_on_sequence(unary_inputs)

    inspectors_cplx_real = unary_dryrun(bbpt_complex_real)

    inspectors_cplx_imag = unary_dryrun(bbpt_complex_imag)

    inspectors_cplx_conjugate = unary_dryrun(bbpt_complex_conjugate)

    inspectors_cplx_norm_squared = unary_dryrun(bbpt_complex_norm_squared)

    for i in range(N):
        binary_args = binary_inputs[i]
        x, y = tuple(map(pytuple_to_complex, binary_args))

        unary_args = unary_inputs[i]
        u = pytuple_to_complex(unary_args[0])

        # Binary:

        inspector_cplx_add = inspectors_cplx_add[i]

        inspector_cplx_mult = inspectors_cplx_mult[i]

        # Unary:

        inspector_cplx_real = inspectors_cplx_real[i]

        inspector_cplx_imag = inspectors_cplx_imag[i]

        inspector_cplx_conjugate = inspectors_cplx_conjugate[i]

        inspector_cplx_norm_squared = inspectors_cplx_norm_squared[i]

        assert x + y == pytuple_to_complex(
            inspector_cplx_add.last_log()
        ), inspector_cplx_add.report(binary_args, f"failed for {binary_args}", row=i)

        assert x * y == pytuple_to_complex(
            inspector_cplx_mult.last_log()
        ), inspector_cplx_mult.report(binary_args, f"failed for {binary_args}", row=i)

        assert u.real == pytuple_to_int(
            inspector_cplx_real.last_log()
        ), inspector_cplx_real.report(unary_args, f"failed for {unary_args}", row=i)

        assert u.imag == pytuple_to_int(
            inspector_cplx_imag.last_log()
        ), inspector_cplx_imag.report(unary_args, f"failed for {unary_args}", row=i)

        assert u.conjugate() == pytuple_to_complex(
            inspector_cplx_conjugate.last_log()
        ), inspector_cplx_conjugate.report(
            unary_args, f"failed for {unary_args}", row=i
        )

        assert u * u.conjugate() == pytuple_to_int(
            inspector_cplx_norm_squared.last_log()
        ), inspector_cplx_norm_squared.report(
            unary_args, f"failed for {unary_args}", row=i
        )


def py_factorial(n):
    return 1 if n <= 1 else n * py_factorial(n - 1)


def test_conditional_factorial():
    ptdre = PyTealDryRunExecutor(conditional_factorial, pt.Mode.Application)
    inputs = [(n,) for n in range(20)]
    inspectors = ptdre.dryrun_on_sequence(inputs)
    for i, args in enumerate(inputs):
        inspector = inspectors[i]
        n = args[0]
        assert inspector.passed(), inspector.report(args, row=i + 1)

        expected = py_factorial(n)
        assert expected == inspector.last_log(), inspector.report(args, row=i + 1)

    n = 21
    args = (n,)
    inspector = ptdre.dryrun(args)
    assert inspector.rejected(), inspector.report(
        args, f"FAILED: should have rejected for {n=}", row=n + 1
    )
    assert inspector.error(), inspector.report(
        args, f"FAILED: should error for {n=}", row=n + 1
    )



from typing import Optional, cast
from pyteal.ast.app import OnComplete
from pyteal.errors import TealInputError, TealTypeError
from pyteal.ast.while_ import While
from pyteal.ast.expr import Expr
from pyteal.ast.global_ import Global
from pyteal.ast.seq import Seq
from pyteal.ast.int import Int
from pyteal.ast.bytes import Bytes
from pyteal.ast.itxn import InnerTxnBuilder
from pyteal.ast.scratchvar import ScratchVar
from pyteal.ast.txn import TxnField, TxnType
from pyteal.ast.for_ import For
from pyteal.types import TealType, require_type
from enum import Enum

ON_CALL_APP = Bytes("base16", "068101")  # v6 program "int 1"


class OpUpMode(Enum):
    """An Enum object that defines the mode used for the OpUp utility.

    Note: the Explicit mode requires the app id to be provided
    through the foreign apps array in order for it to be accessible
    during evaluation.
    """

    #: The app to call must be provided by the user.
    Explicit = 0

    #: The app to call is created then deleted for each request to increase budget.
    OnCall = 1


class OpUpFeeSource(Enum):
    """An Enum object that defines the source for fees for the OpUp utility."""

    #: Only the excess fee (credit) on the outer group should be used (set inner_tx.fee=0)
    GroupCredit = 0
    #: The app's account will cover all fees (set inner_tx.fee=Global.min_tx_fee())
    AppAccount = 1
    #: First the excess will be used, remaining fees will be taken from the app account
    Any = 2


def _fee_by_source(source: OpUpFeeSource) -> Optional[Expr]:
    match source:
        case OpUpFeeSource.GroupCredit:
            return Int(0)
        case OpUpFeeSource.AppAccount:
            return Global.min_txn_fee()
        case _:
            return None


class OpUp:
    """Utility for increasing opcode budget during app execution.

    Requires program version 6 or higher.

    Example:
        .. code-block:: python

            # OnCall mode: doesn't accept target_app_id as an argument
            opup = OpUp(OpUpMode.OnCall)
            program_with_opup = Seq(
                ...,
                opup.ensure_budget(Int(1000)),
                ...,
            )

            # Explicit mode: requires target_app_id as an argument
            opup = OpUp(OpUpMode.Explicit, Int(1))
            program_with_opup = Seq(
                ...,
                opup.ensure_budget(Int(1000)),
                ...,
            )
    """

    def __init__(self, mode: OpUpMode, target_app_id: Expr = None):
        """Create a new OpUp object.

        Args:
            mode: OpUpMode that determines the style of budget increase
                to use. See the OpUpMode Enum for more information.
            target_app_id (optional): In Explicit mode, the OpUp utility
                requires the app_id to target for inner app calls. Defaults
                to None.
        """

        # With only OnCall and Explicit modes supported, the mode argument
        # isn't strictly necessary, but it will most likely be required if
        # we do decide to add more modes in the future.
        if mode == OpUpMode.Explicit:
            if target_app_id is None:
                raise TealInputError(
                    "target_app_id must be specified in Explicit OpUp mode"
                )
            require_type(target_app_id, TealType.uint64)
        elif mode == OpUpMode.OnCall:
            if target_app_id is not None:
                raise TealInputError("target_app_id is not used in OnCall OpUp mode")
        else:
            raise TealInputError("Invalid OpUp mode provided")

        self.mode = mode
        self.target_app_id = target_app_id

    def _construct_itxn(self, inner_fee: Optional[Expr]) -> Expr:
        fields: dict[TxnField, Expr | list[Expr]] = {
            TxnField.type_enum: TxnType.ApplicationCall
        }

        # If an inner_fee is specified
        # add it to the transaction fields
        if inner_fee is not None:
            require_type(inner_fee, TealType.uint64)
            fields[TxnField.fee] = inner_fee

        if self.mode == OpUpMode.Explicit:
            fields |= {TxnField.application_id: cast(Expr, self.target_app_id)}
        else:
            fields |= {
                TxnField.on_completion: OnComplete.DeleteApplication,
                TxnField.approval_program: ON_CALL_APP,
                TxnField.clear_state_program: ON_CALL_APP,
            }

        return InnerTxnBuilder.Execute(fields)

    def ensure_budget(
        self, required_budget: Expr, fee_source: OpUpFeeSource = OpUpFeeSource.Any
    ) -> Expr:
        """Ensure that the budget will be at least the required_budget.

        Args:
            required_budget: minimum op-code budget to ensure for the
                upcoming execution.
            fee_source (optional): source that should be used for covering fees on
                the inner transactions that are generated.

        Note: the available budget just prior to calling ensure_budget() must be
        high enough to execute the budget increase code. The exact budget required
        depends on the provided required_budget expression, but a budget of ~20
        should be sufficient for most use cases. If lack of budget is an issue then
        consider moving the call to ensure_budget() earlier in the pyteal program."""
        require_type(required_budget, TealType.uint64)
        if not type(fee_source) == OpUpFeeSource:
            raise TealTypeError(type(fee_source), OpUpFeeSource)

        # A budget buffer is necessary to deal with an edge case of ensure_budget():
        #   if the current budget is equal to or only slightly higher than the
        #   required budget then it's possible for ensure_budget() to return with a
        #   current budget less than the required budget. The buffer prevents this
        #   from being the case.
        buffer = Int(10)
        buffered_budget = ScratchVar(TealType.uint64)
        return Seq(
            buffered_budget.store(required_budget + buffer),
            While(buffered_budget.load() > Global.opcode_budget()).Do(
                self._construct_itxn(inner_fee=_fee_by_source(fee_source))
            ),
        )

    def maximize_budget(
        self, fee: Expr, fee_source: OpUpFeeSource = OpUpFeeSource.Any
    ) -> Expr:
        """Maximize the available opcode budget without spending more than the given fee.

        Args:
            fee: fee expenditure cap for the op-code budget maximization.
            fee_source (optional): source that should be used for covering fees on
                the inner transactions that are generated.

        Note: the available budget just prior to calling maximize_budget() must be
        high enough to execute the budget increase code. The exact budget required
        depends on the provided fee expression, but a budget of ~25 should be
        sufficient for most use cases. If lack of budget is an issue then consider
        moving the call to maximize_budget() earlier in the pyteal program."""
        require_type(fee, TealType.uint64)
        if not type(fee_source) == OpUpFeeSource:
            raise TealTypeError(type(fee_source), OpUpFeeSource)

        i = ScratchVar(TealType.uint64)
        n = fee / Global.min_txn_fee()
        return For(i.store(Int(0)), i.load() < n, i.store(i.load() + Int(1))).Do(
            self._construct_itxn(inner_fee=_fee_by_source(fee_source))
        )


OpUp.__module__ = "pyteal"
